; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"compact flash procedures for ARCADEmini firmware boot code"

	include		"P18F8520.INC"

	include		"IoPins.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		A1_REG, B1_REG, A2_REG, B2_REG, A4_REG
	extern		CF_SECTOR_CNT



;===========================================================================================
b_prog		CODE			; boot code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; wait for compact flash card to get ready
; output:	STATUS,C	set on error (card gone, timeout)
; changes:	B2_REG
;-------------------------------------------------------------------------------------------
B_CF_WAIT_READY:
; initialize counter
	setf	B2_REG+1		; 65535 times with 10 cycles each
	setf	B2_REG
B_CF_WAIT_READY_WAIT:
	clrwdt
; return error if card is gone
	bsf	STATUS,C		; set error (for return statement)
	btfsc	PIN_CF_nCD1		; skip if card is inserted
	return
; end loop if card is ready
	bcf	STATUS,C		; clear error (for return statement)
	btfsc	PIN_CF_RDY		; skip if card is not ready
	return
; loop
	decfsz	B2_REG
	bra	B_CF_WAIT_READY_WAIT
	decfsz	B2_REG+1
	bra	B_CF_WAIT_READY_WAIT
; timeout
	bsf	STATUS,C		; set error (for return statement)
	return

;-------------------------------------------------------------------------------------------
; read a compact flash register
; input:	W		register address (0..7)
; output:	A1_REG		value read from register
;		STATUS,C	set on error
; changes:	B2_REG, WREG
;-------------------------------------------------------------------------------------------
B_CF_READ_REG:
; wait for compact flash card to get ready
	call	B_CF_WAIT_READY
	btfsc	STATUS,C		; skip if successful
	return				; return with error flag set
; read register
	movwf	PORT_SH_A		; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nOE		; set output enable
	nop				; (CF needs max. 383 us to provide data)
	nop
	nop
	nop
	movff	PORT_SH_D,A1_REG	; read data
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; read a compact flash register multiple times
; input:	W		register address (0..7)
;		FSR0		address of memory where to store data
;		A2_REG		number of bytes to read
; output:	STATUS,C	set on error
; changes:	FSR0, A2_REG, B2_REG, B1_REG, WREG
;-------------------------------------------------------------------------------------------
B_CF_READ_REG_EX:
; read register multiple times
	tstfsz	A2_REG			; increment A2_REG+1 if A2_REG is not 0
	incf	A2_REG+1		; to make "double decfsz" work
	movwf	B1_REG			; save register address in B1_REG
B_CF_READ_REG_EX_LOOP:
; wait for compact flash card to get ready
	call	B_CF_WAIT_READY
	btfsc	STATUS,C		; skip if successful
	return				; return with error flag set
; read register
	movff	B1_REG,PORT_SH_A	; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nOE		; set output enable
	nop				; (CF needs max. 383 us to provide data)
	nop
	nop
	nop
	movff	PORT_SH_D,POSTINC0	; read data into buffer
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; loop
	decfsz	A2_REG			; count down number of bytes in A2_REG
	bra	B_CF_READ_REG_EX_LOOP
	decfsz	A2_REG+1
	bra	B_CF_READ_REG_EX_LOOP
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; dummy read a compact flash register multiple times
; input:	W		register address (0..7)
;		A2_REG		number of bytes to read
; output:	STATUS,C	set on error
; changes:	A2_REG, B2_REG, B1_REG, WREG
;-------------------------------------------------------------------------------------------
B_CF_READ_REG_DUMMY:
; dummy read register multiple times
	tstfsz	A2_REG			; increment A2_REG+1 if A2_REG is not 0
	incf	A2_REG+1		; to make "double decfsz" work
	movwf	B1_REG			; save register address in B1_REG
B_CF_READ_REG_DUMMY_LOOP:
; wait for compact flash card to get ready
	call	B_CF_WAIT_READY
	btfsc	STATUS,C		; skip if successful
	return				; return with error flag set
; dummy read register
	movff	B1_REG,PORT_SH_A	; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nOE		; set output enable
	nop				; (CF needs max. 383 us to provide data)
	nop
	nop
	nop				; do not read data, this is a dummy read
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; loop
	decfsz	A2_REG			; count down number of bytes in A2_REG
	bra	B_CF_READ_REG_DUMMY_LOOP
	decfsz	A2_REG+1
	bra	B_CF_READ_REG_DUMMY_LOOP
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; write a compact flash register
; input:	W		register address (0..7)
;		A1_REG		value to write into register
; output:	STATUS,C	set on error
; changes:	B2_REG, WREG
;-------------------------------------------------------------------------------------------
B_CF_WRITE_REG:
; wait for compact flash card to get ready
	call	B_CF_WAIT_READY
	btfsc	STATUS,C		; skip if successful
	return				; return with error flag set
; write register
	movwf	PORT_SH_A		; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nWE		; set write enable
	movff	A1_REG,PORT_SH_D	; write data
	movlw	PORT_SH_D_TR_O		; set data port to output
	movwf	PORT_SH_D_TRIS
	nop				; (write enable must be active for 385 us)
	nop
	bsf	PIN_SH_nWE		; clear write enable
	movlw	PORT_SH_D_TR_I		; set data port back to input
	movwf	PORT_SH_D_TRIS
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; identify compact compact flash
; - verifies that LBA is supported
; - vefifies that sector size is at least 512 bytes
; output:	CF_SECTOR_CNT	number of sectors
;		STATUS,C	set on error
; changes:	FSR0, A1_REG, A2_REG, B2_REG, WREG
;-------------------------------------------------------------------------------------------
B_CF_IDENTIFY:
; load command register
	movlw	0xEC			; 0xEC = identify drive
	movwf	A1_REG
	movlw	7			; 7 (write) = command register
	call	B_CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; read status
	movlw	7			; 7 (read) = status register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
; check status
	movf	A1_REG,W
	andlw	0xF3			; check that BUSY(D7)=0, RDY(D6)=1, DWF(D5)=0, DSC(D4)=1,
	xorlw	0x50			; DRQ(D3)=X, CORR(D2)=X, IDX(D1)=0, ERR(D0)=0
	bsf	STATUS,C		; set error (for return statement)
	btfss	STATUS,Z		; exit with error if status is not okay
	return
; check compact flash id low (byte 0)
	movlw	0			; 0 = data register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
	bsf	STATUS,C		; set error (for return statements)
	movf	A1_REG,W		; expect 0x8A
	xorlw	0x8A
	btfss	STATUS,Z		; return error if id does not match
	return
; check compact flash id high (byte 1)
	movlw	0			; 0 = data register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
	bsf	STATUS,C		; set error (for return statements)
	movf	A1_REG,W		; expect 0x84
	xorlw	0x84
	btfss	STATUS,Z		; return error if id does not match
	return
; dummy read bytes (2..98)
	movlw	97			; read 97 bytes
	movwf	A2_REG
	movlw	0
	movwf	A2_REG+1
	movlw	0			; 0 = data register
	call	B_CF_READ_REG_DUMMY
	btfsc	STATUS,C		; exit in case of error
	return
; check if LBA is supported (byte 99, bit 1)
	movlw	0			; 0 = data register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
	bsf	STATUS,C		; set error (for return statements)
	movf	A1_REG,W		; expect bit 1 to be set
	andlw	0x02
	btfsc	STATUS,Z		; return error if LBA is not supported
	return
; dummy read bytes (100..119)
	movlw	20			; read 20 bytes
	movwf	A2_REG
	movlw	0
	movwf	A2_REG+1
	movlw	0			; 0 = data register
	call	B_CF_READ_REG_DUMMY
	btfsc	STATUS,C		; exit in case of error
	return
; get number of sectors in LBA mode (bytes 120..123)
	movlw	0			; 0 = data register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A1_REG,CF_SECTOR_CNT	; save byte 0
	movlw	0			; 0 = data register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A1_REG,CF_SECTOR_CNT+1	; save byte 1
	movlw	0			; 0 = data register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A1_REG,CF_SECTOR_CNT+2	; save byte 2
	movlw	0			; 0 = data register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A1_REG,CF_SECTOR_CNT+3	; save byte 3
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; read a compact flash sector in LBA mode
; input:	A4_REG		number of sector to read
;		FSR0		address of memory where to store data (512 bytes)
; output:	STATUS,C	set on error
; changes:	FSR0, A1_REG, A2_REG, B2_REG, WREG
;-------------------------------------------------------------------------------------------
B_CF_READ_SECTOR:
; load sector number into registers
	movf	A4_REG+3,W		; bits 24..27 of sector number
	andlw	0x0F
	iorlw	0xE0			; drive 0, LBA mode
	movwf	A1_REG
	movlw	6			; 6 = head register
	call	B_CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+2,A1_REG		; bits 16..23 of sector number
	movlw	5			; 5 = high part of cylinder register
	call	B_CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+1,A1_REG		; bits 8..15 of sector number
	movlw	4			; 4 = low part of cylinder register
	call	B_CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG,A1_REG		; bits 0..7 of sector number
	movlw	3			; 3 = sector number register
	call	B_CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movlw	1			; sector count = 1
	movwf	A1_REG
	movlw	2			; 2 = sector count register
	call	B_CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; load command register
	movlw	0x20			; 0x20 = read sector
	movwf	A1_REG
	movlw	7			; 7 (write) = command register
	call	B_CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; read status
	movlw	7			; 7 (read) = status register
	call	B_CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	return
; check status
	movf	A1_REG,W
	andlw	0xF3			; check that BUSY(D7)=0, RDY(D6)=1, DWF(D5)=0,
	xorlw	0x50			; DSC(D4)=1, DRQ(3)=X, CORR(D2)=X, IDX(D1)=0, ERR(D0)=0
	bsf	STATUS,C		; set error (for return statement)
	btfss	STATUS,Z		; exit with error if status is not okay
	return
; read data
	clrf	A2_REG			; read 512 bytes
	movlw	0x02
	movwf	A2_REG+1
	movlw	0			; 0 = data register
	call	B_CF_READ_REG_EX	; (FSR0 is set by caller)
	btfsc	STATUS,C		; exit in case of error
	return
; return success
	bcf	STATUS,C
	return



;===========================================================================================
; global procedures
;===========================================================================================
	global		B_CF_IDENTIFY, B_CF_READ_SECTOR



	end
