; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"TEA2 procedures for ARCADEmini firmware boot code"

	include		"P18F8520.INC"

	include		"IoPins.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		A1_REG, B1_REG, C1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG
	extern		A4_REG, B4_REG, C4_REG
	extern		BUFFER_OV



;===========================================================================================
b_prog		CODE			; boot code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; TEA 2.0 block decryption
; input:	BUFFER_OV+0..15		key (16 bytes = 128 bit)
;		A4_REG			1st part of block to decrypt
;		B4_REG			2nd part of block to decrypt
; output:	A4_REG			1st part of decrypted block
;		B4_REG			2nd part of decrypted block
; changes:	FSR0, WREG
;		A1_REG, B1_REG
;		A2_REG, B2_REG, C2_REG, D2_REG
;		C4_REG
;-------------------------------------------------------------------------------------------
B_TEA2_BLK_DEC:
; y = A4_REG, z = B4_REG, sum = C4_REG, key = BUFFER_OV, delta = 0x9E3779B9
; sum = 32 * delta
	movlw	0x20
	movwf	C4_REG+0
	movlw	0x37
	movwf	C4_REG+1
	movlw	0xEF
	movwf	C4_REG+2
	movlw	0xC6
	movwf	C4_REG+3
; 32 rounds
	movlw	32
	movwf	A1_REG
B_TEA2_BLK_DEC_ROUND:
; (y << 4 ^ y >> 5) + y
	movff	A4_REG+0,A2_REG+0		; B2_REG:A2_REG = y << 4
	movff	A4_REG+1,A2_REG+1
	movff	A4_REG+2,B2_REG+0
	movff	A4_REG+3,B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	movff	A4_REG+0,C2_REG+0		; D2_REG:C2_REG = y >> 5
	movff	A4_REG+1,C2_REG+1
	movff	A4_REG+2,D2_REG+0
	movff	A4_REG+3,D2_REG+1
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	movf	C2_REG+0,W			; B2_REG:A2_REG = (y << 4 ^ y >> 5)
	xorwf	A2_REG+0
	movf	C2_REG+1,W
	xorwf	A2_REG+1
	movf	D2_REG+0,W
	xorwf	B2_REG+0
	movf	D2_REG+1,W
	xorwf	B2_REG+1
	movf	A4_REG+0,W			; B2_REG:A2_REG = (y << 4 ^ y >> 5) + y
	addwf	A2_REG+0
	movf	A4_REG+1,W
	addwfc	A2_REG+1
	movf	A4_REG+2,W
	addwfc	B2_REG+0
	movf	A4_REG+3,W
	addwfc	B2_REG+1
; sum + key[sum >> 11 & 3]
	movf	C4_REG+1,W			; B1_REG = (sum >> 11 & 3) << 3
	andlw	0x18
	movwf	B1_REG
	lfsr	FSR0,BUFFER_OV			; FSR0 = &key[sum >> 11 & 3]
	rrncf	B1_REG,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movf	C4_REG+0,W			; D2_REG:C2_REG = sum + key[sum >> 11 & 3]
	addwf	POSTINC0,W
	movwf	C2_REG+0
	movf	C4_REG+1,W
	addwfc	POSTINC0,W
	movwf	C2_REG+1
	movf	C4_REG+2,W
	addwfc	POSTINC0,W
	movwf	D2_REG+0
	movf	C4_REG+3,W
	addwfc	POSTINC0,W
	movwf	D2_REG+1
; z -= (y << 4 ^ y >> 5) + y ^ sum + key[sum >> 11 & 3]
	movf	A2_REG+0,W
	xorwf	C2_REG+0,W
	subwf	B4_REG+0
	movf	A2_REG+1,W
	xorwf	C2_REG+1,W
	subwfb	B4_REG+1
	movf	B2_REG+0,W
	xorwf	D2_REG+0,W
	subwfb	B4_REG+2
	movf	B2_REG+1,W
	xorwf	D2_REG+1,W
	subwfb	B4_REG+3
; sum -= delta
	movlw	0xB9
	subwf	C4_REG+0
	movlw	0x79
	subwfb	C4_REG+1
	movlw	0x37
	subwfb	C4_REG+2
	movlw	0x9E
	subwfb	C4_REG+3
; (z << 4 ^ z >> 5) + z
	movff	B4_REG+0,A2_REG+0		; B2_REG:A2_REG = z << 4
	movff	B4_REG+1,A2_REG+1
	movff	B4_REG+2,B2_REG+0
	movff	B4_REG+3,B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	bcf	STATUS,C
	rlcf	A2_REG+0
	rlcf	A2_REG+1
	rlcf	B2_REG+0
	rlcf	B2_REG+1
	movff	B4_REG+0,C2_REG+0		; D2_REG:C2_REG = z >> 5
	movff	B4_REG+1,C2_REG+1
	movff	B4_REG+2,D2_REG+0
	movff	B4_REG+3,D2_REG+1
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	bcf	STATUS,C
	rrcf	D2_REG+1
	rrcf	D2_REG+0
	rrcf	C2_REG+1
	rrcf	C2_REG+0
	movf	C2_REG+0,W			; B2_REG:A2_REG = (z << 4 ^ z >> 5)
	xorwf	A2_REG+0
	movf	C2_REG+1,W
	xorwf	A2_REG+1
	movf	D2_REG+0,W
	xorwf	B2_REG+0
	movf	D2_REG+1,W
	xorwf	B2_REG+1
	movf	B4_REG+0,W			; B2_REG:A2_REG = (z << 4 ^ z >> 5) + z
	addwf	A2_REG+0
	movf	B4_REG+1,W
	addwfc	A2_REG+1
	movf	B4_REG+2,W
	addwfc	B2_REG+0
	movf	B4_REG+3,W
	addwfc	B2_REG+1
; sum + key[sum & 3]
	movf	C4_REG,W			; B1_REG = sum & 3
	andlw	0x03
	movwf	B1_REG
	lfsr	FSR0,BUFFER_OV			; FSR0 = &key[sum & 3]
	rlncf	B1_REG
	rlncf	B1_REG,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movf	C4_REG+0,W			; D2_REG:C2_REG = sum + key[sum & 3]
	addwf	POSTINC0,W
	movwf	C2_REG+0
	movf	C4_REG+1,W
	addwfc	POSTINC0,W
	movwf	C2_REG+1
	movf	C4_REG+2,W
	addwfc	POSTINC0,W
	movwf	D2_REG+0
	movf	C4_REG+3,W
	addwfc	POSTINC0,W
	movwf	D2_REG+1
; y -= (z << 4 ^ z >> 5) + z ^ sum + key[sum & 3]
	movf	A2_REG+0,W
	xorwf	C2_REG+0,W
	subwf	A4_REG+0
	movf	A2_REG+1,W
	xorwf	C2_REG+1,W
	subwfb	A4_REG+1
	movf	B2_REG+0,W
	xorwf	D2_REG+0,W
	subwfb	A4_REG+2
	movf	B2_REG+1,W
	xorwf	D2_REG+1,W
	subwfb	A4_REG+3
; next round
	decfsz	A1_REG
	bra	B_TEA2_BLK_DEC_ROUND
	return



;-------------------------------------------------------------------------------------------
; TEA 2.0 decryption
; input:	BUFFER_OV+0..15		key (16 bytes = 128 bit)
;		BUFFER_OV+16..23	previous encrypted block (8 bytes = 64 bits)
;					(zero for first block)
;		FSR1			pointer to encrypted data
;					(decrypted data will be written here, too)
;		C1_REG			number of blocks to decrypt
;					(one block = 8 bytes = 64 bits)
; output:	BUFFER_OV+16..23	last encrypted block (8 bytes = 64 bits)
;					(used as previous encrypted block in next call)
; changes:	FSR0, FSR1, WREG
;		A1_REG, B1_REG, C1_REG
;		A2_REG, B2_REG, C2_REG, D2_REG
;		A4_REG, B4_REG, C4_REG
;-------------------------------------------------------------------------------------------
B_TEA2_DEC:
; prev = BUFFER_OV+16..23
B_TEA2_DEC_ROUND:
	clrwdt
; load block
	movff	POSTINC1,A4_REG+0
	movff	POSTINC1,A4_REG+1
	movff	POSTINC1,A4_REG+2
	movff	POSTINC1,A4_REG+3
	movff	POSTINC1,B4_REG+0
	movff	POSTINC1,B4_REG+1
	movff	POSTINC1,B4_REG+2
	movff	INDF1,B4_REG+3
; decrypt block
	call	B_TEA2_BLK_DEC
; xor result with previous encrypted block
	movff	BUFFER_OV+16,WREG
	xorwf	A4_REG+0
	movff	BUFFER_OV+17,WREG
	xorwf	A4_REG+1
	movff	BUFFER_OV+18,WREG
	xorwf	A4_REG+2
	movff	BUFFER_OV+19,WREG
	xorwf	A4_REG+3
	movff	BUFFER_OV+20,WREG
	xorwf	B4_REG+0
	movff	BUFFER_OV+21,WREG
	xorwf	B4_REG+1
	movff	BUFFER_OV+22,WREG
	xorwf	B4_REG+2
	movff	BUFFER_OV+23,WREG
	xorwf	B4_REG+3
; remember last encrypted block
	movff	POSTDEC1,BUFFER_OV+23
	movff	POSTDEC1,BUFFER_OV+22
	movff	POSTDEC1,BUFFER_OV+21
	movff	POSTDEC1,BUFFER_OV+20
	movff	POSTDEC1,BUFFER_OV+19
	movff	POSTDEC1,BUFFER_OV+18
	movff	POSTDEC1,BUFFER_OV+17
	movff	INDF1,BUFFER_OV+16
; write decrypted block
	movff	A4_REG+0,POSTINC1
	movff	A4_REG+1,POSTINC1
	movff	A4_REG+2,POSTINC1
	movff	A4_REG+3,POSTINC1
	movff	B4_REG+0,POSTINC1
	movff	B4_REG+1,POSTINC1
	movff	B4_REG+2,POSTINC1
	movff	B4_REG+3,POSTINC1
; next block
	decfsz	C1_REG
	bra	B_TEA2_DEC_ROUND
	return



;===========================================================================================
; global procedures
;===========================================================================================
	global		B_TEA2_DEC



	end
