; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"compact flash procedures for ARCADEmini firmware"

	include		"P18F8520.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0, BITS1
	extern		A1_REG, B1_REG, C1_REG, A2_REG, B2_REG, C2_REG, A4_REG



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

; detect counter for compact flash card
; - updated in timer 4 interrupt
; - incremented if compact flash detected and smaller than DETECT_CF_MAX
; - set to 0x00 if no compact flash detected
; - different actions are done when reaching different values
CF_DETECT	res	1
CF_DETECT_RST_S	equ	20	; set compact flash reset
CF_DETECT_RST_C	equ	30	; clear compactflash reset
CF_DETECT_RDY	equ	70	; wait for compactflash ready
CF_DETECT_DONE	equ	100	; compact flash inserted
CF_DETECT_MAX	equ	100

; state of compact flash machinery
CF_STATE	res	1
CF_STATE_IDLE			equ	0x00	; no action in progress
CF_STATE_IDENTIFY_1		equ	0x10	; CF_IDENTIFY
CF_STATE_IDENTIFY_2		equ	0x11
CF_STATE_IDENTIFY_3		equ	0x12
CF_STATE_READ_SECTOR_1		equ	0x20	; CF_READ_SECTOR
CF_STATE_READ_SECTOR_2		equ	0x21
CF_STATE_READ_SECTOR_PART_1	equ	0x30	; CF_READ_SECTOR_PART
CF_STATE_READ_SECTOR_PART_2	equ	0x31
CF_STATE_READ_SECTOR_PART_3	equ	0x32
CF_STATE_WRITE_SECTOR_1		equ	0x40	; CF_WRITE_SECTOR
CF_STATE_WRITE_SECTOR_2		equ	0x41
CF_STATE_WRITE_SECTOR_3		equ	0x42

; timeout counter
CF_TIMEOUT	res	1
CF_TIMEOUT_VAL	equ	250		; 250 times CF_TASK (0.25ms) is timeout (62.5ms)

; variables to store parameters of actions in progress
CF_BUF_ADDR	res	2		; buffer address
CF_OFFSET	res	2		; offset into CF sector
CF_NO_BYTES	res	2		; number of bytes to read/write

; number of sectors on the compact flash (little endian)
; - == 0 if no compact flash is inserted
; - != 0 if compact flash is inserted
CF_SECTOR_CNT	res	4



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; read a compact flash register
; input:	W		register address (0..7)
; output:	A1_REG		value read from register
;		STATUS,C	set on error
; changes:	WREG
;-------------------------------------------------------------------------------------------
CF_READ_REG:
; check that compact flash card is inserted and ready
	bsf	STATUS,C		; set error (for return statement)
	btfsc	PIN_CF_nCD1		; skip if card is inserted
	return
	btfss	PIN_CF_RDY		; skip if card is ready
	return
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; read register
	movwf	PORT_SH_A		; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nOE		; set output enable
	nop				; (CF needs max. 383 us to provide data)
	nop
	nop
	nop
	movff	PORT_SH_D,A1_REG	; read data
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; read a compact flash register multiple times
; input:	W		register address (0..7)
;		FSR0		address of memory where to store data
;		A1_REG		number of bytes to read
; output:	STATUS,C	set on error
; changes:	FSR0, A1_REG, B1_REG, WREG
;-------------------------------------------------------------------------------------------
CF_READ_REG_EX:
	movwf	B1_REG			; save register address in B1_REG
CF_READ_REG_EX_LOOP:
; check that compact flash card is inserted and ready
	bsf	STATUS,C		; set error (for return statement)
	btfsc	PIN_CF_nCD1		; skip if card is inserted
	return
	btfss	PIN_CF_RDY		; skip if card is ready
	return
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; read register
	movff	B1_REG,PORT_SH_A	; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nOE		; set output enable
	nop				; (CF needs max. 383 us to provide data)
	nop
	nop
	nop
	movff	PORT_SH_D,POSTINC0	; read data into buffer
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; loop
	decfsz	A1_REG			; count down number of bytes in A1_REG
	bra	CF_READ_REG_EX_LOOP
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; dummy read a compact flash register multiple times
; input:	W		register address (0..7)
;		A1_REG		number of bytes to read
; output:	STATUS,C	set on error
; changes:	A1_REG, B1_REG, WREG
;-------------------------------------------------------------------------------------------
CF_READ_REG_DUMMY:
	movwf	B1_REG			; save register address in B1_REG
CF_READ_REG_DUMMY_LOOP:
; check that compact flash card is inserted and ready
	bsf	STATUS,C		; set error (for return statement)
	btfsc	PIN_CF_nCD1		; skip if card is inserted
	return
	btfss	PIN_CF_RDY		; skip if card is ready
	return
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; dummy read register
	movff	B1_REG,PORT_SH_A	; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nOE		; set output enable
	nop				; (CF needs max. 383 us to provide data)
	nop
	nop
	nop				; do not read data, this is a dummy read
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; loop
	decfsz	A1_REG			; count down number of bytes in A1_REG
	bra	CF_READ_REG_DUMMY_LOOP
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; write a compact flash register
; input:	W		register address (0..7)
;		A1_REG		value to write into register
; output:	STATUS,C	set on error
; changes:	WREG
;-------------------------------------------------------------------------------------------
CF_WRITE_REG:
; check that compact flash card is inserted and ready
	bsf	STATUS,C		; set error (for return statement)
	btfsc	PIN_CF_nCD1		; skip if card is inserted
	return
	btfss	PIN_CF_RDY		; skip if card is ready
	return
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; write register
	movwf	PORT_SH_A		; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nWE		; set write enable
	movff	A1_REG,PORT_SH_D	; write data
	movlw	PORT_SH_D_TR_O		; set data port to output
	movwf	PORT_SH_D_TRIS
	nop				; (write enable must be active for 385 us)
	nop
	bsf	PIN_SH_nWE		; clear write enable
	movlw	PORT_SH_D_TR_I		; set data port back to input
	movwf	PORT_SH_D_TRIS
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; write a compact flash register multiple times
; input:	W		register address (0..7)
;		FSR0		address of data
;		A1_REG		number of bytes to write
; output:	STATUS,C	set on error
; changes:	FSR0, A1_REG, B1_REG, WREG
;-------------------------------------------------------------------------------------------
CF_WRITE_REG_EX:
	movwf	B1_REG			; save register address in B1_REG
CF_WRITE_REG_EX_LOOP:
; check that compact flash card is inserted and ready
	bsf	STATUS,C		; set error (for return statement)
	btfsc	PIN_CF_nCD1		; skip if card is inserted
	return
	btfss	PIN_CF_RDY		; skip if card is ready
	return
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; write register
	movff	B1_REG,PORT_SH_A	; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nWE		; set write enable
	movff	POSTINC0,PORT_SH_D	; write data
	movlw	PORT_SH_D_TR_O		; set data port to output
	movwf	PORT_SH_D_TRIS
	nop				; (write enable must be active for 385 us)
	nop
	bsf	PIN_SH_nWE		; clear write enable
	movlw	PORT_SH_D_TR_I		; set data port back to input
	movwf	PORT_SH_D_TRIS
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; loop
	decfsz	A1_REG			; count down number of bytes in A1_REG
	bra	CF_WRITE_REG_EX_LOOP
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; dummy write a compact flash register multiple times (with data 0x00)
; input:	W		register address (0..7)
;		A1_REG		number of bytes to write
; output:	STATUS,C	set on error
; changes:	A1_REG, B1_REG, WREG
;-------------------------------------------------------------------------------------------
CF_WRITE_REG_DUMMY:
	movwf	B1_REG			; save register address in B1_REG
CF_WRITE_REG_DUMMY_LOOP:
; check that compact flash card is inserted and ready
	bsf	STATUS,C		; set error (for return statement)
	btfsc	PIN_CF_nCD1		; skip if card is inserted
	return
	btfss	PIN_CF_RDY		; skip if card is ready
	return
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; write register
	movff	B1_REG,PORT_SH_A	; output register address
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_CF_nCE		; set chip enable
	bcf	PIN_SH_nWE		; set write enable
	clrf	PORT_SH_D		; write 0x00 as data
	movlw	PORT_SH_D_TR_O		; set data port to output
	movwf	PORT_SH_D_TRIS
	nop				; (write enable must be active for 385 us)
	nop
	bsf	PIN_SH_nWE		; clear write enable
	movlw	PORT_SH_D_TR_I		; set data port back to input
	movwf	PORT_SH_D_TRIS
	bsf	PIN_CF_nCE		; clear chip enable
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; loop
	decfsz	A1_REG			; count down number of bytes in A1_REG
	bra	CF_WRITE_REG_DUMMY_LOOP
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; compact flash initialization
; changes:	-
;-------------------------------------------------------------------------------------------
CF_INIT:
; initialize compact flash state
	clrf	CF_DETECT		; detect counter
	movlw	CF_STATE_IDLE		; no action in progress
	movwf	CF_STATE
; clear number of sectors
	clrf	CF_SECTOR_CNT+0
	clrf	CF_SECTOR_CNT+1
	clrf	CF_SECTOR_CNT+2
	clrf	CF_SECTOR_CNT+3
; initialize global bits
	bcf	BIT_CF_BUSY		; not busy
	bcf	BIT_CF_ERROR		; no error
; done
	return

;-------------------------------------------------------------------------------------------
; compact flash task
; - called periodically
; changes:	WREG, STATUS
;-------------------------------------------------------------------------------------------
CF_TASK:

; compact flash detection
	btfss	PIN_CF_nCD1		; jump if compact flash detected
	bra	CF_TASK_DETECT
; no compact flash detected
	movf	CF_DETECT,W		; jump to end if detect counter is 0
	bz	CF_TASK_DETECT_END
	clrf	CF_DETECT		; reset detect counter
	movlw	CF_STATE_IDLE		; no action in progress
	movwf	CF_STATE
	clrf	CF_SECTOR_CNT+0		; clear sector count
	clrf	CF_SECTOR_CNT+1
	clrf	CF_SECTOR_CNT+2
	clrf	CF_SECTOR_CNT+3
	bcf	BIT_CF_ERROR		; no error
	bcf	BIT_CF_BUSY		; not busy
	bra	CF_TASK_DETECT_END
; compact flash detected
CF_TASK_DETECT:
	movlw	CF_DETECT_MAX		; jump to end if detect counter is at maximum value
	cpfslt	CF_DETECT
	bra	CF_TASK_DETECT_END
	incf	CF_DETECT		; increment detect counter
	movlw	CF_DETECT_RST_S		; if detect counter is at reset-set value, ...
	xorwf	CF_DETECT,W
	btfsc	STATUS,Z
	bsf	PIN_CF_RESET		; ... set compact flash reset bit
	movlw	CF_DETECT_RST_C		; if detect counter is at reset-clear value, ...
	xorwf	CF_DETECT,W
	btfsc	STATUS,Z
	bcf	PIN_CF_RESET		; ... clear compact flash reset bit
	movlw	CF_DETECT_RDY		; if detect counter is at wait-for-ready value, ...
	xorwf	CF_DETECT,W
	bnz	CF_TASK_DETECT_RDY_END
	btfss	PIN_CF_RDY		; ... and if compact flash is not yet ready, ...
	decf	CF_DETECT		; ... decrement detect counter
CF_TASK_DETECT_RDY_END:
	movlw	CF_DETECT_DONE		; if detect counter is at done value, ...
	xorwf	CF_DETECT,W
	bnz	CF_TASK_DETECT_DONE_END
	movlw	CF_STATE_IDLE		; .... set state to no action in progress
	movwf	CF_STATE
	clrf	CF_SECTOR_CNT+0		; ... clear sector count
	clrf	CF_SECTOR_CNT+1
	clrf	CF_SECTOR_CNT+2
	clrf	CF_SECTOR_CNT+3
	bcf	BIT_CF_ERROR		; ... set to no error
	bcf	BIT_CF_BUSY		; ... set to not busy
	call	CF_IDENTIFY		; ... start compact flash identification
CF_TASK_DETECT_DONE_END:
CF_TASK_DETECT_END:

; exit task if state is idle
	movlw	CF_STATE_IDLE
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	return

; jump to task procedure according to state
	movlw	CF_STATE_IDENTIFY_1
	cpfslt	CF_STATE
	bra	CF_TASK_STATE_IDENTIFY
	bra	CF_TASK_UNKNOWN
CF_TASK_STATE_IDENTIFY:
	movlw	CF_STATE_READ_SECTOR_1
	cpfslt	CF_STATE
	bra	CF_TASK_STATE_READ_SECTOR
	movlw	CF_STATE_IDENTIFY_1
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_IDENTIFY_1
	movlw	CF_STATE_IDENTIFY_2
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_IDENTIFY_2
	movlw	CF_STATE_IDENTIFY_3
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_IDENTIFY_3
	bra	CF_TASK_UNKNOWN
CF_TASK_STATE_READ_SECTOR:
	movlw	CF_STATE_READ_SECTOR_PART_1
	cpfslt	CF_STATE
	bra	CF_TASK_STATE_READ_SECTOR_PART
	movlw	CF_STATE_READ_SECTOR_1
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_READ_SECTOR_1
	movlw	CF_STATE_READ_SECTOR_2
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_READ_SECTOR_2
	bra	CF_TASK_UNKNOWN
CF_TASK_STATE_READ_SECTOR_PART:
	movlw	CF_STATE_WRITE_SECTOR_1
	cpfslt	CF_STATE
	bra	CF_TASK_STATE_WRITE_SECTOR
	movlw	CF_STATE_READ_SECTOR_PART_1
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_READ_SECTOR_PART_1
	movlw	CF_STATE_READ_SECTOR_PART_2
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_READ_SECTOR_PART_2
	movlw	CF_STATE_READ_SECTOR_PART_3
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_READ_SECTOR_PART_3
	bra	CF_TASK_UNKNOWN
CF_TASK_STATE_WRITE_SECTOR:
	movlw	CF_STATE_WRITE_SECTOR_1
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_WRITE_SECTOR_1
	movlw	CF_STATE_WRITE_SECTOR_2
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_WRITE_SECTOR_2
	movlw	CF_STATE_WRITE_SECTOR_3
	xorwf	CF_STATE,W
	btfsc	STATUS,Z
	bra	CF_TASK_WRITE_SECTOR_3
CF_TASK_UNKNOWN:
; set state to idle
	movlw	CF_STATE_IDLE
	movwf	CF_STATE
; done
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_IDENTIFY - 1st part
;-------------------------------------------------------------------------------------------
CF_TASK_IDENTIFY_1:
; wait for compact flash to get ready
	incf	CF_TIMEOUT		; check for timeout
	movlw	CF_TIMEOUT_VAL
	cpfslt	CF_TIMEOUT
	bra	CF_TASK_ERROR
	btfss	PIN_CF_RDY		; exit if card not ready
	return
; read status
	movlw	7			; 7 (read) = status register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; check status
	movf	A1_REG,W
	andlw	0xF3			; check that BUSY(D7)=0, RDY(D6)=1, DWF(D5)=0, DSC(D4)=1,
	xorlw	0x50			; DRQ(D3)=X, CORR(D2)=X, IDX(D1)=0, ERR(D0)=0
	btfss	STATUS,Z		; exit with error if status is not okay
	bra	CF_TASK_ERROR
; check compact flash id low (byte 0)
	movlw	0			; 0 = data register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
	movf	A1_REG,W		; expect 0x8A
	xorlw	0x8A
	btfss	STATUS,Z		; return error if id does not match
	bra	CF_TASK_ERROR
; check compact flash id high (byte 1)
	movlw	0			; 0 = data register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
	movf	A1_REG,W		; expect 0x84
	xorlw	0x84
	btfss	STATUS,Z		; return error if id does not match
	bra	CF_TASK_ERROR
; dummy read bytes (2..49)
	movlw	48			; read 48 bytes
	movwf	A1_REG
	movlw	0			; 0 = data register
	call	CF_READ_REG_DUMMY
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; set state to identifying compact flash - 2nd part
	movlw	CF_STATE_IDENTIFY_2
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_IDENTIFY - 2nd part
;-------------------------------------------------------------------------------------------
CF_TASK_IDENTIFY_2:
; dummy read bytes (50..98)
	movlw	49			; read 49 bytes
	movwf	A1_REG
	movlw	0			; 0 = data register
	call	CF_READ_REG_DUMMY
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; check if LBA is supported (byte 99, bit 1)
	movlw	0			; 0 = data register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
	movf	A1_REG,W		; expect bit 1 to be set
	andlw	0x02
	btfsc	STATUS,Z		; return error if LBA is not supported
	bra	CF_TASK_ERROR
; set state to identifying compact flash - 3rd part
	movlw	CF_STATE_IDENTIFY_3
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_IDENTIFY - 3rd part
;-------------------------------------------------------------------------------------------
CF_TASK_IDENTIFY_3:
; dummy read bytes (100..119)
	movlw	20			; read 20 bytes
	movwf	A1_REG
	movlw	0			; 0 = data register
	call	CF_READ_REG_DUMMY
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; get number of sectors in LBA mode (bytes 120..123) into A4_REG
	movlw	0			; 0 = data register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
	movff	A1_REG,A4_REG+0		; save byte 0
	movlw	0			; 0 = data register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
	movff	A1_REG,A4_REG+1		; save byte 1
	movlw	0			; 0 = data register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
	movff	A1_REG,A4_REG+2		; save byte 2
	movlw	0			; 0 = data register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
	movff	A1_REG,A4_REG+3		; save byte 3
; set number of sectors
	movff	A4_REG+0,CF_SECTOR_CNT+0
	movff	A4_REG+1,CF_SECTOR_CNT+1
	movff	A4_REG+2,CF_SECTOR_CNT+2
	movff	A4_REG+3,CF_SECTOR_CNT+3
; set state to idle
	movlw	CF_STATE_IDLE
	movwf	CF_STATE
; return success
	bcf	BIT_CF_ERROR		; clear error bit
	bcf	BIT_CF_BUSY		; clear busy bit
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_READ_SECTOR - 1st part
;-------------------------------------------------------------------------------------------
CF_TASK_READ_SECTOR_1:
; wait for compact flash to get ready
	incf	CF_TIMEOUT		; check for timeout
	movlw	CF_TIMEOUT_VAL
	cpfslt	CF_TIMEOUT
	bra	CF_TASK_ERROR
	btfss	PIN_CF_RDY		; exit if card not ready
	return
; read status
	movlw	7			; 7 (read) = status register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; check status
	movf	A1_REG,W
	andlw	0xF3			; check that BUSY(D7)=0, RDY(D6)=1, DWF(D5)=0,
	xorlw	0x50			; DSC(D4)=1, DRQ(3)=X, CORR(D2)=X, IDX(D1)=0, ERR(D0)=0
	btfss	STATUS,Z		; exit with error if status is not okay
	bra	CF_TASK_ERROR
; set number of bytes to read to 0x200
	movlw	0x02
	movwf	CF_NO_BYTES+1
	clrf	CF_NO_BYTES+0
; set state to reading compact flash sector - 2nd part
	movlw	CF_STATE_READ_SECTOR_2
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_READ_SECTOR - 2nd part
;-------------------------------------------------------------------------------------------
CF_TASK_READ_SECTOR_2:
; get number of bytes to read this time and subtract it from number of bytes to read at all
	movlw	0x40			; prepare to read maximum number of bytes (0x40)
	movwf	A1_REG
	movf	CF_NO_BYTES+1,W		; if >= 0x100 bytes, read maximum number of bytes
	bnz	CF_TASK_READ_SECTOR_2_MAX
	movlw	0x40			; if >= 0x40 bytes, read maximum number of bytes	
	cpfslt	CF_NO_BYTES+0
	bra	CF_TASK_READ_SECTOR_2_MAX
	movff	CF_NO_BYTES+0,A1_REG	; read CF_NO_BYTES bytes (here always < 0x40)
CF_TASK_READ_SECTOR_2_MAX:
	movf	A1_REG,W		; subtract number of bytes to read now
	subwf	CF_NO_BYTES+0
	movlw	0
	subwfb	CF_NO_BYTES+1
; get address of buffer and advance it by number of bytes to read
	movff	CF_BUF_ADDR+0,FSR0L
	movff	CF_BUF_ADDR+1,FSR0H
	movf	A1_REG,W		; advance address
	addwf	CF_BUF_ADDR+0
	movlw	0
	addwfc	CF_BUF_ADDR+1
; read data
	movlw	0			; 0 = data register
	call	CF_READ_REG_EX
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; done in this step, if still bytes to read
	movf	CF_NO_BYTES+0,W
	iorwf	CF_NO_BYTES+1,W
	btfss	STATUS,Z
	return
; return success
	bcf	BIT_CF_ERROR		; clear error bit
	bcf	BIT_CF_BUSY		; clear busy bit
	movlw	CF_STATE_IDLE		; set state to idle
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_READ_SECTOR_PART - 1st part
;-------------------------------------------------------------------------------------------
CF_TASK_READ_SECTOR_PART_1:
; wait for compact flash to get ready
	incf	CF_TIMEOUT		; check for timeout
	movlw	CF_TIMEOUT_VAL
	cpfslt	CF_TIMEOUT
	bra	CF_TASK_ERROR
	btfss	PIN_CF_RDY		; exit if card not ready
	return
; read status
	movlw	7			; 7 (read) = status register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; check status
	movf	A1_REG,W
	andlw	0xF3			; check that BUSY(D7)=0, RDY(D6)=1, DWF(D5)=0,
	xorlw	0x50			; DSC(D4)=1, DRQ(3)=X, CORR(D2)=X, IDX(D1)=0, ERR(D0)=0
	btfss	STATUS,Z		; exit with error if status is not okay
	bra	CF_TASK_ERROR
; set state to reading part of compact flash sector - 2nd part
	movlw	CF_STATE_READ_SECTOR_PART_2
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_READ_SECTOR_PART - 2nd part
;-------------------------------------------------------------------------------------------
CF_TASK_READ_SECTOR_PART_2:
; skip this step if nothing to read
	movf	CF_OFFSET+0,W
	iorwf	CF_OFFSET+1,W
	bz	CF_TASK_READ_SECTOR_PART_2_SKIP
; get number of bytes to read this time and subtract it from number of bytes to read at all
	movlw	0x40			; prepare to read maximum number of bytes (0x40)
	movwf	A1_REG
	movf	CF_OFFSET+1,W		; if >= 0x100 bytes, read maximum number of bytes
	bnz	CF_TASK_READ_SECTOR_PART_2_MAX
	movlw	0x40			; if >= 0x40 bytes, read maximum number of bytes	
	cpfslt	CF_OFFSET+0
	bra	CF_TASK_READ_SECTOR_PART_2_MAX
	movff	CF_OFFSET+0,A1_REG	; read CF_OFFSET bytes (here always < 0x40)
CF_TASK_READ_SECTOR_PART_2_MAX:
	movf	A1_REG,W		; subtract number of bytes to read now
	subwf	CF_OFFSET+0
	movlw	0
	subwfb	CF_OFFSET+1
; dummy read data
	movlw	0			; 0 = data register
	call	CF_READ_REG_DUMMY
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; done in this step, if still bytes to dummy read
	movf	CF_OFFSET+0,W
	iorwf	CF_OFFSET+1,W
	btfss	STATUS,Z
	return
CF_TASK_READ_SECTOR_PART_2_SKIP:
; set state to reading part of compact flash sector - 3rd part
	movlw	CF_STATE_READ_SECTOR_PART_3
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_READ_SECTOR_PART - 3rd part
;-------------------------------------------------------------------------------------------
CF_TASK_READ_SECTOR_PART_3:
; get number of bytes to read this time and subtract it from number of bytes to read at all
	movlw	0x40			; prepare to read maximum number of bytes (0x40)
	movwf	A1_REG
	movf	CF_NO_BYTES+1,W		; if >= 0x100 bytes, read maximum number of bytes
	bnz	CF_TASK_READ_SECTOR_PART_3_MAX
	movlw	0x40			; if >= 0x40 bytes, read maximum number of bytes	
	cpfslt	CF_NO_BYTES+0
	bra	CF_TASK_READ_SECTOR_PART_3_MAX
	movff	CF_NO_BYTES+0,A1_REG	; read CF_NO_BYTES bytes (here always < 0x40)
CF_TASK_READ_SECTOR_PART_3_MAX:
	movf	A1_REG,W		; subtract number of bytes to read now
	subwf	CF_NO_BYTES+0
	movlw	0
	subwfb	CF_NO_BYTES+1
; get address of buffer and advance it by number of bytes to read
	movff	CF_BUF_ADDR+0,FSR0L
	movff	CF_BUF_ADDR+1,FSR0H
	movf	A1_REG,W		; advance address
	addwf	CF_BUF_ADDR+0
	movlw	0
	addwfc	CF_BUF_ADDR+1
; read data
	movlw	0			; 0 = data register
	call	CF_READ_REG_EX
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; done in this step, if still bytes to read
	movf	CF_NO_BYTES+0,W
	iorwf	CF_NO_BYTES+1,W
	btfss	STATUS,Z
	return
; return success
	bcf	BIT_CF_ERROR		; clear error bit
	bcf	BIT_CF_BUSY		; clear busy bit
	movlw	CF_STATE_IDLE		; set state to idle
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_WRITE_SECTOR - 1st part
;-------------------------------------------------------------------------------------------
CF_TASK_WRITE_SECTOR_1:
; wait for compact flash to get ready
	incf	CF_TIMEOUT		; check for timeout
	movlw	CF_TIMEOUT_VAL
	cpfslt	CF_TIMEOUT
	bra	CF_TASK_ERROR
	btfss	PIN_CF_RDY		; exit if card not ready
	return
; read status
	movlw	7			; 7 (read) = status register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; check status
	movf	A1_REG,W
	andlw	0xF3			; check that BUSY(D7)=0, RDY(D6)=1, DWF(D5)=0,
	xorlw	0x50			; DSC(D4)=1, DRQ(3)=X, CORR(D2)=X, IDX(D1)=0, ERR(D0)=0
	bsf	STATUS,C		; set error (for return statement)
	btfss	STATUS,Z		; exit with error if status is not okay
	bra	CF_TASK_ERROR
; set number of bytes to write to 0x200
	movlw	0x02
	movwf	CF_NO_BYTES+1
	clrf	CF_NO_BYTES+0
; set state to writing compact flash sector - 2nd part
	movlw	CF_STATE_WRITE_SECTOR_2
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_WRITE_SECTOR - 2nd part
;-------------------------------------------------------------------------------------------
CF_TASK_WRITE_SECTOR_2:
; get number of bytes to write this time and subtract it from number of bytes to write at all
	movlw	0x40			; prepare to write maximum number of bytes (0x40)
	movwf	A1_REG
	movf	CF_NO_BYTES+1,W		; if >= 0x100 bytes, write maximum number of bytes
	bnz	CF_TASK_WRITE_SECTOR_2_MAX
	movlw	0x40			; if >= 0x40 bytes, write maximum number of bytes	
	cpfslt	CF_NO_BYTES+0
	bra	CF_TASK_WRITE_SECTOR_2_MAX
	movff	CF_NO_BYTES+0,A1_REG	; write CF_NO_BYTES bytes (here always < 0x40)
CF_TASK_WRITE_SECTOR_2_MAX:
	movf	A1_REG,W		; subtract number of bytes to write now
	subwf	CF_NO_BYTES+0
	movlw	0
	subwfb	CF_NO_BYTES+1
; get address of buffer and advance it by number of bytes to write
	movff	CF_BUF_ADDR+0,FSR0L
	movff	CF_BUF_ADDR+1,FSR0H
	movf	A1_REG,W		; advance address
	addwf	CF_BUF_ADDR+0
	movlw	0
	addwfc	CF_BUF_ADDR+1
; write data
	movlw	0			; 0 = data register
	call	CF_WRITE_REG_EX
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; done in this step, if still bytes to write
	movf	CF_NO_BYTES+0,W
	iorwf	CF_NO_BYTES+1,W
	btfss	STATUS,Z
	return
; clear timeout counter
	clrf	CF_TIMEOUT
; set state to writing compact flash sector - 3rd part
	movlw	CF_STATE_WRITE_SECTOR_3
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; task procedure of CF_WRITE_SECTOR - 3rd part
;-------------------------------------------------------------------------------------------
CF_TASK_WRITE_SECTOR_3:
; wait for compact flash to get ready
	incf	CF_TIMEOUT		; check for timeout
	movlw	CF_TIMEOUT_VAL
	cpfslt	CF_TIMEOUT
	bra	CF_TASK_ERROR
	btfss	PIN_CF_RDY		; exit if card not ready
	return
; read status
	movlw	7			; 7 (read) = status register
	call	CF_READ_REG
	btfsc	STATUS,C		; exit in case of error
	bra	CF_TASK_ERROR
; check status
	movf	A1_REG,W
	andlw	0xF3			; check that BUSY(D7)=0, RDY(D6)=1, DWF(D5)=0,
	xorlw	0x50			; DSC(D4)=1, DRQ(3)=X, CORR(D2)=X, IDX(D1)=0, ERR(D0)=0
	bsf	STATUS,C		; set error (for return statement)
	btfss	STATUS,Z		; exit with error if status is not okay
	bra	CF_TASK_ERROR
; return success
	bcf	BIT_CF_ERROR		; clear error bit
	bcf	BIT_CF_BUSY		; clear busy bit
	movlw	CF_STATE_IDLE		; set state to idle
	movwf	CF_STATE
	return

;-------------------------------------------------------------------------------------------
; error during compact flash task
;-------------------------------------------------------------------------------------------
CF_TASK_ERROR:
; set error bit
	bsf	BIT_CF_ERROR
; clear busy bit
	bcf	BIT_CF_BUSY
; set state to idle
	movlw	CF_STATE_IDLE
	movwf	CF_STATE
; done
	return

;-------------------------------------------------------------------------------------------
; start compact compact flash identification
; - verifies that LBA is supported
; - verifies that sector size is at least 512 bytes
; output:	CF_SECTOR_CNT	number of sectors
;		STATUS,C	set on error
; changes:	FSR0, A1_REG, A2_REG, WREG
; on success, the requested action is started
; - BIT_CF_BUSY is set
; - BIT_CF_BUSY will be cleared if requested action was done
; - BIT_CF_ERROR will be set if an error occured
;-------------------------------------------------------------------------------------------
CF_IDENTIFY:
; check that compact flash is idle
	bsf	STATUS,C		; set error bit (for return statement)
	movlw	CF_STATE_IDLE
	cpfseq	CF_STATE		; skip if idle
	return
; load command register
	movlw	0xEC			; 0xEC = identify drive
	movwf	A1_REG
	movlw	7			; 7 (write) = command register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; clear timeout counter
	clrf	CF_TIMEOUT
; set state to identifying compact flash - 1 st part
	movlw	CF_STATE_IDENTIFY_1
	movwf	CF_STATE
; set busy bit
	bsf	BIT_CF_BUSY
; return success (action was started)
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; start reading a compact flash sector in LBA mode
; input:	A4_REG		number of sector to read
;		FSR0		address of memory where to store data (512 bytes)
; output:	STATUS,C	set on error
; changes:	A1_REG, A2_REG, WREG
; on success, the requested action is started
; - BIT_CF_BUSY is set
; - BIT_CF_BUSY will be cleared if requested action was done
; - BIT_CF_ERROR will be set if an error occured
;-------------------------------------------------------------------------------------------
CF_READ_SECTOR:
; check that compact flash is idle
	bsf	STATUS,C		; set error bit (for return statement)
	movlw	CF_STATE_IDLE
	cpfseq	CF_STATE		; skip if idle
	return
; load sector number into registers
	movf	A4_REG+3,W		; bits 24..27 of sector number
	andlw	0x0F
	iorlw	0xE0			; drive 0, LBA mode
	movwf	A1_REG
	movlw	6			; 6 = head register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+2,A1_REG		; bits 16..23 of sector number
	movlw	5			; 5 = high part of cylinder register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+1,A1_REG		; bits 8..15 of sector number
	movlw	4			; 4 = low part of cylinder register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG,A1_REG		; bits 0..7 of sector number
	movlw	3			; 3 = sector number register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movlw	1			; sector count = 1
	movwf	A1_REG
	movlw	2			; 2 = sector count register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; load command register
	movlw	0x20			; 0x20 = read sector
	movwf	A1_REG
	movlw	7			; 7 (write) = command register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; remember buffer address
	movff	FSR0L,CF_BUF_ADDR+0
	movff	FSR0H,CF_BUF_ADDR+1
; clear timeout counter
	clrf	CF_TIMEOUT
; set state to reading compact flash sector - 1st part
	movlw	CF_STATE_READ_SECTOR_1
	movwf	CF_STATE
; set busy bit
	bsf	BIT_CF_BUSY
; return success (action was started)
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; start reading a part of a compact flash sector in LBA mode
; input:	A4_REG		number of sector to read
;		A2_REG		offset where to start
;		FSR0		address of memory where to store data
;		C2_REG		number of bytes to read
; output:	STATUS,C	set on error
; changes:	A1_REG, WREG
; on success, the requested action is started
; - BIT_CF_BUSY is set
; - BIT_CF_BUSY will be cleared if requested action was done
; - BIT_CF_ERROR will be set if an error occured
;-------------------------------------------------------------------------------------------
CF_READ_SECTOR_PART:
; check that compact flash is idle
	bsf	STATUS,C		; set error bit (for return statement)
	movlw	CF_STATE_IDLE
	cpfseq	CF_STATE		; skip if idle
	return
; load sector number into registers
	movf	A4_REG+3,W		; bits 24..27 of sector number
	andlw	0x0F
	iorlw	0xE0			; drive 0, LBA mode
	movwf	A1_REG
	movlw	6			; 6 = head register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+2,A1_REG		; bits 16..23 of sector number
	movlw	5			; 5 = high part of cylinder register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+1,A1_REG		; bits 8..15 of sector number
	movlw	4			; 4 = low part of cylinder register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG,A1_REG		; bits 0..7 of sector number
	movlw	3			; 3 = sector number register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movlw	1			; sector count = 1
	movwf	A1_REG
	movlw	2			; 2 = sector count register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; load command register
	movlw	0x20			; 0x20 = read sector
	movwf	A1_REG
	movlw	7			; 7 (write) = command register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; remember buffer address, offset and number of bytes
	movff	FSR0L,CF_BUF_ADDR+0
	movff	FSR0H,CF_BUF_ADDR+1
	movff	A2_REG+0,CF_OFFSET+0
	movff	A2_REG+1,CF_OFFSET+1
	movff	C2_REG+0,CF_NO_BYTES+0
	movff	C2_REG+1,CF_NO_BYTES+1
; clear timeout counter
	clrf	CF_TIMEOUT
; set state to reading part of compact flash sector - 1st part
	movlw	CF_STATE_READ_SECTOR_PART_1
	movwf	CF_STATE
; set busy bit
	bsf	BIT_CF_BUSY
; return success (action was started)
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; start writing a compact flash sector in LBA mode
; input:	A4_REG		number of sector to read
;		FSR0		address of memory from where to read data (512 bytes)
; output:	STATUS,C	set on error
; changes:	A1_REG, WREG
; on success, the requested action is started
; - BIT_CF_BUSY is set
; - BIT_CF_BUSY will be cleared if requested action was done
; - BIT_CF_ERROR will be set if an error occured
;-------------------------------------------------------------------------------------------
CF_WRITE_SECTOR:
; check that compact flash is idle
	bsf	STATUS,C		; set error bit (for return statement)
	movlw	CF_STATE_IDLE
	cpfseq	CF_STATE		; skip if idle
	return
; load sector number into registers
	movf	A4_REG+3,W		; bits 24..27 of sector number
	andlw	0x0F
	iorlw	0xE0			; drive 0, LBA mode
	movwf	A1_REG
	movlw	6			; 6 = head register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+2,A1_REG		; bits 16..23 of sector number
	movlw	5			; 5 = high part of cylinder register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG+1,A1_REG		; bits 8..15 of sector number
	movlw	4			; 4 = low part of cylinder register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movff	A4_REG,A1_REG		; bits 0..7 of sector number
	movlw	3			; 3 = sector number register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
	movlw	1			; sector count = 1
	movwf	A1_REG
	movlw	2			; 2 = sector count register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; load command register
	movlw	0x30			; 0x30 = write sector
	movwf	A1_REG
	movlw	7			; 7 (write) = command register
	call	CF_WRITE_REG
	btfsc	STATUS,C		; exit in case of error
	return
; remember buffer address
	movff	FSR0L,CF_BUF_ADDR+0
	movff	FSR0H,CF_BUF_ADDR+1
; clear timeout counter
	clrf	CF_TIMEOUT
; set state to writing compact flash sector - 1st part
	movlw	CF_STATE_WRITE_SECTOR_1
	movwf	CF_STATE
; set busy bit
	bsf	BIT_CF_BUSY
; return success (action was started)
	bcf	STATUS,C
	return



;===========================================================================================
; global variables
;===========================================================================================
	global		CF_SECTOR_CNT



;===========================================================================================
; global procedures
;===========================================================================================
	global		CF_INIT, CF_TASK
	global		CF_IDENTIFY, CF_READ_SECTOR, CF_READ_SECTOR_PART, CF_WRITE_SECTOR



	end
