; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"ARCADEmini firmware"

	include		"P18F8520.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Constants.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		CMD_SECTOR, CMD_OFFSET
	extern		CF_SECTOR_CNT



;===========================================================================================
; imported procedures
;===========================================================================================
	extern		IF_READ_SECTOR
	extern		CF_READ_SECTOR
	extern		O_INIT, O_INT_TIMER2
	extern		USB_INIT, USB_TASK
	extern		CMD_INIT, CMD_EXEC
	extern		CF_INIT, CF_TASK
	extern		EE_INIT, EE_CONFIG, EE_TASK
	extern		EE_SET_OUT_STYLE, EE_RESET_OUT_TIMING, EE_RESET_OUT_GREYMAP



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

; storage for global bits (see Bits.inc)
BITS0		res	1
BITS1		res	1

; general purpose registers
A1_REG		res	1
B1_REG		res	1
C1_REG		res	1
D1_REG		res	1
A2_REG		res	2
B2_REG		res	2
C2_REG		res	2
D2_REG		res	2
E2_REG		res	2
A4_REG		res	4
B4_REG		res	4
C4_REG		res	4

; delay variables
; - decremented in timer 4 interrupt if BIT_DELAY is set
; - BIT_DELAY is cleared if DELAY reaches 0
DELAY		res	2

; timer variables
; - updated in timer 4 interrupt
TIME_TICKS	res	1	; timer ticks
TIME_MSEC	res	1	; milliseconds (0..9)
TIME_SEC100	res	1	; hundredth seconds (0..99)
TIME_SEC	res	1	; seconds (0..59)
TIME_MIN	res	1	; minutes (0..59)
TIME_HOUR	res	1	; hours (0..23)

; last value of pseudorandom number generator (PRNG)
RANDOM		res	1
; help valiable for randomizing the PRNG in interrupt 4
RANDOM_HELP	res	1

; state of keys
; - updated in timer 4 interrupt
KEYS_CUR	res	1	; current state of keys
KEYS_OLD1	res	1	; state of keys last time
KEYS_OLD2	res	1	; state of keys the time before last time
; debounced state of keys
; - updated in timer 4 interrupt
KEYS		res	1	; key state (to work with)
KEYS_PREV	res	1	; previos key state
KEYS_PRESSED	res	1	; keys that have been pressed
KEYS_RELEASED	res	1	; keys that have been released
; debounced state of SNES keys
; - updated in timer 4 interrupt
SNES0		res	2	; current key state
SNES1		res	2
SNES0_PREV	res	2	; previous key state
SNES1_PREV	res	2
SNES0_PRESSED	res	2	; keys that have been pressed
SNES1_PRESSED	res	2
SNES0_RELEASED	res	2	; keys that have been released
SNES1_RELEASED	res	2
SNES_CNT	res	1	; counter for shifting in snes keys
; configuration via keys
; - hold keys 2, 3 during power up
; - then press key 5, release key 2, release key 3
; - then press a key (1..4,6..7) to change a single configuration setting
; - then release key 5
KEY_CONFIG	res	1	; state of configuration via keys
KEY_CONFIG_DISABLED	equ	0x00	; disabled
KEY_CONFIG_INIT		equ	0x01	; initial counter value
KEY_CONFIG_WAIT 	equ	0x20	; time for key debouncing to recognize key state 0x06
KEY_CONFIG_06_16	equ	0x21	; key state 0x06, waiting for 0x16
KEY_CONFIG_16_14	equ	0x22	; key state 0x16, waiting for 0x14
KEY_CONFIG_14_10	equ	0x23	; key state 0x14, waiting for 0x10
KEY_CONFIG_10_WAIT	equ	0x24	; key state 0x10, waiting for a single key configuration command

; main state machine
STATE		res	1
STATE_IF_EX	equ	0x00	; executing commands from internal flash
STATE_CF_ID	equ	0x01	; checking ID sector of CF card
STATE_CF_EX	equ	0x02	; executing commands from compact flash
STATE_CF_IF_EX	equ	0x03	; CF present but executing command from internal flash



;===========================================================================================
ram_def		UDATA		; default RAM (this page is selected by default)
;===========================================================================================

RAM_DEF		res	0	; used for BANKSEL



;===========================================================================================
ram_main	UDATA			; normal RAM for main
;===========================================================================================

; backup registers for low priority interrupts
INT_WREG	res	1
INT_STATUS	res	1
INT_BSR		res	1

; command buffer
; - is loaded with a "sector" of internal flash
; - is loaded with a sector of compact flash
; - contains commands to execute
BUFFER		res	512
; overflow command buffer
; - is initialized with 0x00 (0x00 = jump to start of next sector)
; - used if some command starts in BUFFER but extends behind its end
;   if this happens the instruction pointer will point somewhere into this buffer
;   the next command is 0x00 so the next sector is loaded
; - must be at least as long as longest command
BUFFER_OV	res	68
BUFFER_END



;===========================================================================================
entry		CODE			; code entry point
;===========================================================================================
ENTRY:
	goto	INIT

;===========================================================================================
int_hi		CODE			; high priority interrupts
;===========================================================================================
HIGH_INTERRUPT:
	btfsc	PIR1,1			; timer 2 interrupt
	goto	O_INT_TIMER2
	reset				; something went completely wrong, so start over

;===========================================================================================
int_lo		CODE			; low priority interrupts
;===========================================================================================
LOW_INTERRUPT:
	btfsc	PIR3,3			; timer 4 interrupt
	goto	INT_TIMER4
	reset				; something went completely wrong, so start over



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; initialization
;-------------------------------------------------------------------------------------------
INIT:
; set default values for output pins
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_SH_nWE		; clear write enable
	bsf	PIN_CF_nCE		; clear chip enable
	bcf	PIN_CF_RESET		; clear reset
	bsf	PIN_USB_nCS		; clear chip select
	bsf	PIN_USB_nRESET		; clear reset
	clrf	PORT_COL0		; turn off all LEDs
	clrf	PORT_COL1
	clrf	PORT_COL2
	clrf	PORT_ROW2_COL3
	clrf	PORT_ROW1
	clrf	PORT_ROW0
; disable all analog inputs
	movlw	b'00001111'
	movwf	ADCON1
; disable comparator inputs
	movlw	b'00000111'
	movwf	CMCON
; define inputs/outputs
	clrf	TRISA
	clrf	TRISB
	clrf	TRISC
	clrf	TRISD
	clrf	TRISE
	clrf	TRISF
	movlw	b'00011111'
	movwf	TRISG
	setf	TRISH
	clrf	TRISJ
; select default ram bank
	BANKSEL	RAM_DEF
; call initialization procedures
	call	O_INIT
	call	USB_INIT
	call	CMD_INIT
	call	CF_INIT
	call	EE_INIT
; initialize variables
	bcf	BIT_DELAY		; delay not active
	bcf	BIT_DELAY_WAIT		; do not wait for delay
	clrf	TIME_TICKS		; timer
	clrf	TIME_MSEC
	clrf	TIME_SEC100
	clrf	TIME_SEC
	clrf	TIME_MIN
	clrf	TIME_HOUR
	bsf	BIT_TASKS_EXEC		; execute tasks as soon as possible
	clrf	KEYS_CUR		; keys
	clrf	KEYS_OLD1
	clrf	KEYS_OLD2
	clrf	KEYS
	clrf	KEYS_PREV
	clrf	KEYS_PRESSED
	clrf	KEYS_RELEASED
	clrf	SNES0			; snes pad buttons
	clrf	SNES0+1
	clrf	SNES1
	clrf	SNES1+1
	clrf	SNES0_PREV
	clrf	SNES0_PREV+1
	clrf	SNES1_PREV
	clrf	SNES1_PREV+1
	clrf	SNES0_PRESSED
	clrf	SNES0_PRESSED+1
	clrf	SNES1_PRESSED
	clrf	SNES1_PRESSED+1
	clrf	SNES0_RELEASED
	clrf	SNES0_RELEASED+1
	clrf	SNES1_RELEASED
	clrf	SNES1_RELEASED+1
	movlw	KEY_CONFIG_INIT		; initialize configuration via keys
	movwf	KEY_CONFIG
; initialize random number generator
	lfsr	FSR0,0x400		; use 256 bytes of ram starting at 0x400
	clrf	A1_REG
RND_INIT:
	movf	POSTINC0,W
	xorwf	RANDOM
	decfsz	A1_REG
	bra	RND_INIT
; disable timer 0
	bcf	T0CON,7
; initialize timer 2
	movlw	b'00000101'		; timer on, prescaler 1:4, postscaler 1:1
	movwf	T2CON
	movlw	74			; 4*75=300 cycles period
	movwf	PR2
; initialize timer 4
	movlw	b'01001100'		; timer on, prescaler 1:1, postscaler 1:10
	movwf	T4CON
	movlw	249			; 250*10=2500 cycles period
	movwf	PR4
; initialize overflow buffer with 0x00
; (0x00 = jump to start of next sector)
	movlw	BUFFER_END-BUFFER_OV	; size of overflow buffer
	movwf	A1_REG
	lfsr	FSR0,BUFFER_OV		; address of buffer
INIT_BUFFER_OV:
	clrf	POSTINC0		; write 0x00 into overflow buffer
	decfsz	A1_REG
	bra	INIT_BUFFER_OV
; get configuration data from EEPROM
	call	EE_CONFIG
; initialize interrupts
	bsf	RCON,7			; enable interrupt priority levels
	bsf	IPR1,1			; timer 2 interrupt to high priority
	bsf	PIE1,1			; enable timer 2 interrupt
	bcf	IPR3,3			; timer 4 interrupt to low priority
	bsf	PIE3,3			; enable timer 4 interrupt
	bsf	INTCON,7		; enable interrupts
	bsf	INTCON,6
; reset watchdog
	clrwdt
; initialization done, start program
	goto	MAIN

;-------------------------------------------------------------------------------------------
; timer 4 interrupt
; - called every 2500 cycles
; - low priority
; - uses interrupt backup registers INT_x WREG, STATUS and BSR
;-------------------------------------------------------------------------------------------
INT_TIMER4:
	bcf	PIR3,3			; clear interrupt flag of timer 4 interrupt
	movff	WREG,INT_WREG		; save WREG, STATUS and BSR
	movff	STATUS,INT_STATUS
	movff	BSR,INT_BSR

; set flag to execute tasks that need periodic processing
; but cannot be done in interrupt
	bsf	BIT_TASKS_EXEC

; only do something else every millisecond
	incf	TIME_TICKS		; ticks
	movlw	3			; > 3?
	cpfsgt	TIME_TICKS
	bra	INT_TIMER4_END
	clrf	TIME_TICKS		; := 0!

; advance timer
	incf	TIME_MSEC		; milliseconds
	movlw	9			; > 9?
	cpfsgt	TIME_MSEC
	bra	INT_TIMER4_TIMER_END
	clrf	TIME_MSEC		; := 0!
	incf	TIME_SEC100		; hundredth seconds
	movlw	99			; > 99?
	cpfsgt	TIME_SEC100
	bra	INT_TIMER4_TIMER_END
	clrf	TIME_SEC100		; := 0!
	incf	TIME_SEC		; seconds
	movlw	59			; > 59?
	cpfsgt	TIME_SEC
	bra	INT_TIMER4_TIMER_END
	clrf	TIME_SEC		; := 0!
	incf	TIME_MIN		; minutes
	movlw	59			; > 59?
	cpfsgt	TIME_MIN
	bra	INT_TIMER4_TIMER_END
	clrf	TIME_MIN		; := 0!
	incf	TIME_HOUR		; seconds
	movlw	23			; > 23?
	cpfsgt	TIME_HOUR
	bra	INT_TIMER4_TIMER_END
	clrf	TIME_HOUR		; := 0!	
INT_TIMER4_TIMER_END:

; decrement delay every millisecond
	btfss	BIT_DELAY
	bra	INT_TIMER4_DELAY_END	; jump if delay not active
	decf	DELAY			; decrement delay counter
	movlw	0
	subwfb	DELAY+1
	movf	DELAY,W			; check if 0
	iorwf	DELAY+1,W
	bnz	INT_TIMER4_DELAY_END	; jump to end if not 0
	bcf	BIT_DELAY		; delay elapsed -> delay not active
	bcf	BIT_DELAY_WAIT		; delay elapsed -> do not wait for delay
INT_TIMER4_DELAY_END:

; clear PRNG help variable
; - bit 0 is set during key reading if a key was just pressed
	clrf	RANDOM_HELP

; keys can be read here
; - keys share data port of compact-flash and usb-chip
; - interrupt is disabled when compact-flash or usb-chip use this port
; read new key state
	comf	PORT_SH_D,W		; keys are active low
	movwf	KEYS_CUR
; save previos debounced key state
	movff	KEYS,KEYS_PREV
; if last 3 key states were 0 then set key to 0
	movf	KEYS_OLD2,W
	iorwf	KEYS_OLD1,W
	iorwf	KEYS_CUR,W
	andwf	KEYS
; if last 3 key states were 1 then set key to 1
	movf	KEYS_OLD2,W
	andwf	KEYS_OLD1,W
	andwf	KEYS_CUR,W
	iorwf	KEYS
; save old key states
	movff	KEYS_OLD1,KEYS_OLD2
	movff	KEYS_CUR,KEYS_OLD1
; accumulate keys that have been pressed
	comf	KEYS_PREV,W		; keys that were 0
	andwf	KEYS,W			; and are 1
	btfss	STATUS,Z		; (set RANDOM_HELP,0
	bsf	RANDOM_HELP,0		;  if a key was pressed)
	iorwf	KEYS_PRESSED
; accumulate keys that have been released
	comf	KEYS,W			; keys that are 0
	andwf	KEYS_PREV,W		; and were 1
	iorwf	KEYS_RELEASED

; read new snes key state (D0=snes clck, D1=snes load)
	movf	TIME_MSEC,W		; only read snes pads every 10 ms
	bnz	INT_TIMER4_SNES_END

; intialize port pins
	setf	PORT_SH_D		; data pins high
	bcf	PORT_SH_D_CLK		; clear clock
	bcf	PORT_SH_D_LOAD		; clear load
	movlw	PORT_SH_D_TR_O		; set data port to output
	movwf	PORT_SH_D_TRIS
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop

; output a load pulse
	bsf	PORT_SH_D_LOAD
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	bcf	PORT_SH_D_LOAD

; backup last key state
	movff	SNES0,SNES0_PREV
	movff	SNES0+1,SNES0_PREV+1
	movff	SNES1,SNES1_PREV
	movff	SNES1+1,SNES1_PREV+1

; shift in 16 buttons
	movlw	16
	movwf	SNES_CNT
INT_TIMER4_SNES_LOOP:
; read next button of SNES pads
	bcf	STATUS,C		; get button of 1st SNES pad into carry
	btfss	PIN_SH_A_SNES0		; (SNES buttons are inverted)
	bsf	STATUS,C
	rlcf	SNES0			; store button
	rlcf	SNES0+1
	bcf	STATUS,C		; get button of 2nd SNES pad into carry
	btfss	PIN_SH_A_SNES1		; (SNES buttons are inverted)
	bsf	STATUS,C
	rlcf	SNES1			; store button
	rlcf	SNES1+1
; output a clock pulse
	bsf	PORT_SH_D_CLK
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	bcf	PORT_SH_D_CLK
	nop
	nop
	nop
	nop
	nop
	nop
; next button
	decfsz	SNES_CNT
	bra	INT_TIMER4_SNES_LOOP

; cleanup port pins
	movlw	PORT_SH_D_TR_I		; set data port back to input
	movwf	PORT_SH_D_TRIS

; accumulate pressed buttons
	comf	SNES0_PREV,W
	andwf	SNES0,W
	btfss	STATUS,Z		; (set RANDOM_HELP,0
	bsf	RANDOM_HELP,0		;  if a key was pressed)
	iorwf	SNES0_PRESSED
	comf	SNES0_PREV+1,W
	andwf	SNES0+1,W
	btfss	STATUS,Z		; (set RANDOM_HELP,0
	bsf	RANDOM_HELP,0		;  if a key was pressed)
	iorwf	SNES0_PRESSED+1
	comf	SNES1_PREV,W
	andwf	SNES1,W
	btfss	STATUS,Z		; (set RANDOM_HELP,0
	bsf	RANDOM_HELP,0		;  if a key was pressed)
	iorwf	SNES1_PRESSED
	comf	SNES1_PREV+1,W
	andwf	SNES1+1,W
	btfss	STATUS,Z		; (set RANDOM_HELP,0
	bsf	RANDOM_HELP,0		;  if a key was pressed)
	iorwf	SNES1_PRESSED+1
; accumulate released buttons
	comf	SNES0,W
	andwf	SNES0_PREV,W
	iorwf	SNES0_RELEASED
	comf	SNES0+1,W
	andwf	SNES0_PREV+1,W
	iorwf	SNES0_RELEASED+1
	comf	SNES1,W
	andwf	SNES1_PREV,W
	iorwf	SNES1_RELEASED
	comf	SNES1+1,W
	andwf	SNES1_PREV+1,W
	iorwf	SNES1_RELEASED+1
INT_TIMER4_SNES_END:

; if a key was just pressed, randomize PRNG
	btfss	RANDOM_HELP,0
	bra	INT_TIMER4_PRNG_END
; add hundredth seconds to RANDOM
	movf	TIME_SEC100,W
	addwf	RANDOM
INT_TIMER4_PRNG_END:

INT_TIMER4_END:
; clean up
	movff	INT_BSR,BSR		; restore WREG, STATUS and BSR
	movff	INT_STATUS,STATUS
	movff	INT_WREG,WREG
	retfie

;-------------------------------------------------------------------------------------------
; compact flash id string (32 bytes)
;-------------------------------------------------------------------------------------------
COMPACT_FLASH_ID:
	db	"ARCADEmini 26x20/16 ambin 1.0   "

;-------------------------------------------------------------------------------------------
; main loop
;-------------------------------------------------------------------------------------------
MAIN:
; set state to executing commands from internal flash
	movlw	STATE_IF_EX
	movwf	STATE
; start executing commands from internal flash
; (start with "jump to start of sector 1")
	bcf	BIT_CMD_CF		; execute commands from internal flash
	clrf	CMD_SECTOR		; sector 0
	clrf	CMD_SECTOR+1
	clrf	CMD_SECTOR+2
	clrf	CMD_SECTOR+3
	movlw	0x02			; set command pointer just behind command buffer
	movwf	CMD_OFFSET+1		; (overflow buffer contains 0x00 = jump to start of next sector)
	clrf	CMD_OFFSET+0
	bcf	BIT_DELAY		; disable current delay
	bcf	BIT_DELAY_WAIT

; main loop
MAIN_LOOP:
; reset watchdog
	clrwdt

; perhaps execute tasks
	btfss	BIT_TASKS_EXEC
	bra	MAIN_TASKS_END
	bcf	BIT_TASKS_EXEC		; clear indicator bit
	call	USB_TASK		; call tasks
	call	CF_TASK
	call	KEY_CONFIG_TASK
	call	EE_TASK
MAIN_TASKS_END:

; process a command
	movlw	STATE_CF_ID		; do not execute commands when checking ID sector of CF
	cpfseq	STATE
	call	CMD_EXEC

; state machinery - executing command from internal flash
	movlw	STATE_IF_EX
	cpfseq	STATE
	bra	MAIN_IF_EX_END
; check if compact flash appears
  	movf	CF_SECTOR_CNT+0,W
  	iorwf	CF_SECTOR_CNT+1,W
  	iorwf	CF_SECTOR_CNT+2,W
  	iorwf	CF_SECTOR_CNT+3,W
  	btfsc	STATUS,Z
  	bra	MAIN_IF_EX_END		; jump if still no CF
; read first sector of compact flash
	clrf	A4_REG			; sector 0
	clrf	A4_REG+1
	clrf	A4_REG+2
	clrf	A4_REG+3
	lfsr	FSR0,BUFFER		; address of buffer (512 bytes)
	call	CF_READ_SECTOR
	btfsc	STATUS,C
  	bra	MAIN_IF_EX_CF_ERR	; jump if CF error
; set state to checking ID sector of CF
	movlw	STATE_CF_ID
	movwf	STATE
	bra	MAIN_IF_EX_END
; set state to CF present but executing commands from internal flash
MAIN_IF_EX_CF_ERR:
	movlw	STATE_CF_IF_EX
	movwf	STATE
MAIN_IF_EX_END:

; state machinery - reading ID sector of CF
	movlw	STATE_CF_ID
	cpfseq	STATE
	bra	MAIN_CF_ID_END
; check if ID sector has been read
  	btfsc	BIT_CF_BUSY
  	bra	MAIN_CF_ID_END		; do nothing if CF still busy
  	btfsc	BIT_CF_ERROR
  	bra	MAIN_CF_ID_ERR
; check id string (first 32 bytes)
	movlw	32			; 32 bytes
	movwf	A1_REG
	lfsr	FSR0,BUFFER		; address of buffer with CF data
	movlw	0x80			; configure for code read
	movwf	EECON1
	movlw	upper COMPACT_FLASH_ID	; address of id string in code
	movwf	TBLPTRU
	movlw	high COMPACT_FLASH_ID
	movwf	TBLPTRH
	movlw	low COMPACT_FLASH_ID
	movwf	TBLPTRL	
MAIN_CF_ID_CHECK:
	movf	POSTINC0,W		; read byte of CF data
	tblrd*+				; read byte of id string in code
	cpfseq	TABLAT			; compare and skip if match
	bra	MAIN_CF_ID_ERR		; ID sector is invalid
	decfsz	A1_REG
	bra	MAIN_CF_ID_CHECK
; set state to executing commands from compact flash
	movlw	STATE_CF_EX
	movwf	STATE
; start executing commands from compact flash
; (start with "jump to start of sector 1")
	bsf	BIT_CMD_CF		; execute commands from compact flash
	clrf	CMD_SECTOR		; sector 0
	clrf	CMD_SECTOR+1
	clrf	CMD_SECTOR+2
	clrf	CMD_SECTOR+3
	movlw	0x02			; set command pointer just behind command buffer
	movwf	CMD_OFFSET+1		; (overflow buffer contains 0x00 = jump to start of next sector)
	clrf	CMD_OFFSET+0
	bcf	BIT_DELAY		; disable current delay
	bcf	BIT_DELAY_WAIT
	bra	MAIN_CF_ID_END
; set state to CF present but executing commands from internal flash
MAIN_CF_ID_ERR:
	movlw	STATE_CF_IF_EX
	movwf	STATE
; reload current sector from internal flash
; (CF ID check has destroyed sector buffer)
	movff	CMD_SECTOR+0,A4_REG+0	; sector number
	movff	CMD_SECTOR+1,A4_REG+1
	movff	CMD_SECTOR+2,A4_REG+2
	movff	CMD_SECTOR+3,A4_REG+3
	lfsr	FSR0,BUFFER		; address of sector buffer
	call	IF_READ_SECTOR
	btfss	STATUS,C
	bra	MAIN_CF_ID_END
; reloading sector from internal flash failed
; start executing commands from internal flash
; (start with "jump to start of sector 1")
	bcf	BIT_CMD_CF		; execute commands from internal flash
	clrf	CMD_SECTOR		; sector 0
	clrf	CMD_SECTOR+1
	clrf	CMD_SECTOR+2
	clrf	CMD_SECTOR+3
	movlw	0x02			; set command pointer just behind command buffer
	movwf	CMD_OFFSET+1		; (overflow buffer contains 0x00 = jump to start of next sector)
	clrf	CMD_OFFSET+0
	bcf	BIT_DELAY		; disable current delay
	bcf	BIT_DELAY_WAIT
MAIN_CF_ID_END:

; state machinery - executing command from compact flash
	movlw	STATE_CF_EX
	cpfseq	STATE
	bra	MAIN_CF_EX_END
; switch back to internal flash if CMD_EXEC reported error
	btfss	BIT_CMD_CF
	bra	MAIN_CF_EX_ERR
; check if compact flash is gone
  	movf	CF_SECTOR_CNT+0,W
  	iorwf	CF_SECTOR_CNT+1,W
  	iorwf	CF_SECTOR_CNT+2,W
  	iorwf	CF_SECTOR_CNT+3,W
  	btfss	STATUS,Z
  	bra	MAIN_CF_EX_END		; jump if CF still there
; set state to executing commands from internal flash
MAIN_CF_EX_ERR:
	movlw	STATE_IF_EX
	movwf	STATE
; start executing commands from internal flash
; (start with "jump to start of sector 1")
	bcf	BIT_CMD_CF		; execute commands from internal flash
	clrf	CMD_SECTOR		; sector 0
	clrf	CMD_SECTOR+1
	clrf	CMD_SECTOR+2
	clrf	CMD_SECTOR+3
	movlw	0x02			; set command pointer just behind command buffer
	movwf	CMD_OFFSET+1		; (overflow buffer contains 0x00 = jump to start of next sector)
	clrf	CMD_OFFSET+0
	bcf	BIT_DELAY		; disable current delay
	bcf	BIT_DELAY_WAIT
MAIN_CF_EX_END:

; state machinery - CF present but executing commands from internal flash
	movlw	STATE_CF_IF_EX
	cpfseq	STATE
	bra	MAIN_CF_IF_EX_END
; check if compact flash is gone
  	movf	CF_SECTOR_CNT+0,W
  	iorwf	CF_SECTOR_CNT+1,W
  	iorwf	CF_SECTOR_CNT+2,W
  	iorwf	CF_SECTOR_CNT+3,W
  	btfss	STATUS,Z
  	bra	MAIN_CF_IF_EX_END	; jump if CF still there
; set state to executing commands from internal flash
	movlw	STATE_IF_EX
	movwf	STATE
MAIN_CF_IF_EX_END:

; go on with main loop
	bra	MAIN_LOOP



;-------------------------------------------------------------------------------------------
; configuration via keys
;-------------------------------------------------------------------------------------------
KEY_CONFIG_TASK:
; disable timer 4 interrupt (keys are read there)
	bcf	PIE3,3
; exit if configuration via keys is already disabled
	movf	KEY_CONFIG,W
	xorlw	KEY_CONFIG_DISABLED
	btfsc	STATUS,Z
	bra	KEY_CONFIG_TASK_END
; time for key debouncing to recognize key state 0x06
	movlw	KEY_CONFIG_WAIT+1
	cpfslt	KEY_CONFIG
	bra	KEY_CONFIG_TASK_0_END
	incf	KEY_CONFIG
	bra	KEY_CONFIG_TASK_END
KEY_CONFIG_TASK_0_END:
; key state 0x06, waiting for 0x16
	movlw	KEY_CONFIG_06_16
	cpfseq	KEY_CONFIG
	bra	KEY_CONFIG_TASK_1_END
	movf	KEYS,W
	xorlw	0x06
	bz	KEY_CONFIG_TASK_END
	movlw	0x16
	cpfseq	KEYS
	bra	KEY_CONFIG_TASK_DISABLE
	movlw	KEY_CONFIG_16_14
	movwf	KEY_CONFIG
KEY_CONFIG_TASK_1_END:
; key state 0x16, waiting for 0x14
	movlw	KEY_CONFIG_16_14
	cpfseq	KEY_CONFIG
	bra	KEY_CONFIG_TASK_2_END
	movf	KEYS,W
	xorlw	0x16
	bz	KEY_CONFIG_TASK_END
	movlw	0x14
	cpfseq	KEYS
	bra	KEY_CONFIG_TASK_DISABLE
	movlw	KEY_CONFIG_14_10
	movwf	KEY_CONFIG
KEY_CONFIG_TASK_2_END:
; key state 0x14, waiting for 0x10
	movlw	KEY_CONFIG_14_10
	cpfseq	KEY_CONFIG
	bra	KEY_CONFIG_TASK_3_END
	movf	KEYS,W
	xorlw	0x14
	bz	KEY_CONFIG_TASK_END
	movlw	0x10
	cpfseq	KEYS
	bra	KEY_CONFIG_TASK_DISABLE
	movlw	KEY_CONFIG_10_WAIT
	movwf	KEY_CONFIG
KEY_CONFIG_TASK_3_END:
; key state 0x10, waiting for key configuration command
	movlw	KEY_CONFIG_10_WAIT
	cpfseq	KEY_CONFIG
	bra	KEY_CONFIG_TASK_DISABLE
	btfss	KEYS,4			; check that key 5 is held down
	bra	KEY_CONFIG_TASK_DISABLE
	btfss	KEYS,0			; key 1: set LED output style
	bra	KEY_CONFIG_TASK_KEY1_END
	movlw	O_STYLE_LED
	movwf	A1_REG
	call	EE_SET_OUT_STYLE
	bra	KEY_CONFIG_TASK_DISABLE
KEY_CONFIG_TASK_KEY1_END:
	btfss	KEYS,1			; key 2: set LATCH output style
	bra	KEY_CONFIG_TASK_KEY2_END
	movlw	O_STYLE_LATCH
	movwf	A1_REG
	call	EE_SET_OUT_STYLE
	bra	KEY_CONFIG_TASK_DISABLE
KEY_CONFIG_TASK_KEY2_END:
	btfss	KEYS,2			; key 3: reset output timing
	bra	KEY_CONFIG_TASK_KEY3_END
	call	EE_RESET_OUT_TIMING
	bra	KEY_CONFIG_TASK_DISABLE
KEY_CONFIG_TASK_KEY3_END:
	btfss	KEYS,3			; key 4: set AMP output style
	bra	KEY_CONFIG_TASK_KEY4_END
	movlw	O_STYLE_AMP
	movwf	A1_REG
	call	EE_SET_OUT_STYLE
	bra	KEY_CONFIG_TASK_DISABLE
KEY_CONFIG_TASK_KEY4_END:
	btfss	KEYS,5			; key 6: reset output greyscale mapping
	bra	KEY_CONFIG_TASK_KEY6_END
	call	EE_RESET_OUT_GREYMAP
	bra	KEY_CONFIG_TASK_DISABLE
KEY_CONFIG_TASK_KEY6_END:
	bra	KEY_CONFIG_TASK_END
; disable configuration via keys
KEY_CONFIG_TASK_DISABLE:
	movlw	KEY_CONFIG_DISABLED
	movwf	KEY_CONFIG
KEY_CONFIG_TASK_END:
; enable timer 4 interrupt
	bsf	PIE3,3
	return



;===========================================================================================
; global procedures
;===========================================================================================

	global		ENTRY, HIGH_INTERRUPT, LOW_INTERRUPT



;===========================================================================================
; global variables
;===========================================================================================
	global		BITS0, BITS1
	global		A1_REG, B1_REG, C1_REG, D1_REG
	global		A2_REG, B2_REG, C2_REG, D2_REG, E2_REG
	global		A4_REG, B4_REG, C4_REG
	global		DELAY
	global		TIME_SEC100, TIME_SEC, TIME_MIN, TIME_HOUR
	global		KEYS, KEYS_PRESSED, KEYS_RELEASED
	global		SNES0, SNES1, SNES0_PREV, SNES1_PREV
	global		SNES0_PRESSED, SNES1_PRESSED, SNES0_RELEASED, SNES1_RELEASED
	global		BUFFER, BUFFER_OV, BUFFER_END
	global		RANDOM



	end
