; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"output procedures for ARCADEmini firmware"

	include		"P18F8520.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Constants.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0, BITS1
	extern		A1_REG, A2_REG



;===========================================================================================
ram_def		UDATA		; default RAM (this page is selected by default)
;===========================================================================================

; output style to use
O_STYLE		res	1	; output style (unknown styles default to same as 0x00)
O_STYLE_LED	equ	0x00	; LED output style (for LED matrix)
O_STYLE_LATCH	equ	0x01	; LATCH output style (for latch circuit)
O_STYLE_AMP	equ	0x02	; AMPLIFIER output style (for amplifier circuit)
; counter for current greyscale value (0..14)
O_GREY		res	1
; value to compare swapped greyscale values to
O_GREY_VAL	res	1
; counter for O_LINEs to output (0..19)
O_LINE		res	1
; variables to store O_LINE bits for led output
O_LINE0		res	1	; bit 0 = O_LINE 0, ..., bit 7 = O_LINE 7
O_LINE1		res	1	; bit 0 = O_LINE 8, ..., bit 7 = O_LINE 15
O_LINE2		res	1	; bit 0 = O_LINE 16, ..., bit 3 = O_LINE 19, bits 4..7 = always 0
; variables to store bits to output to leds
O_LED0		res	1	; bit 0 = column 0, ..., bit 7 = column 7
O_LED1		res	1	; bit 0 = column 8, ..., bit 7 = column 15
O_LED2		res	1	; bit 0 = column 16, ..., bit 7 = coulmn 23
O_LED3		res	1	; bit 4 = column 24, bit 5 = column 25, bits 3..0,6..7 = always 0
; backup register for FSR2
O_FSR_BAK	res	2



;===========================================================================================
ram_out		UDATA			; normal RAM for output
;===========================================================================================

; two buffers with frames
; each byte contains the 4 bit greyscale values for 2 LEDs
; O_FRAMEx+0, 7..4: greyscale value for top left LED
; O_FRAMEx+0, 3..0: greyscale value for 2nd LED in top row
; O_FRAMEx+1, 7..4: greyscale value for 3rd LED in top row
; ...
; O_FRAMEx+12, 7..4: greyscale value for top right LED
; O_FRAMEx+13, 3..0: greyscale value for leftmost LED in 2nd row
; ...
; O_FRAMEx+259, 3..0: greyscale value for bottom right LED
O_FRAME0	res	260
O_FRAME1	res	260

; BIT_O_LOCK_BUF contains the number of the buffer owned by the interrupt
;   if this bit is cleared, O_FRAME0 is owned by the interrupt and writing to it is forbidden
;   if this bit is set, O_FRAME1 is owned by the interrupt and writing to it is forbidden
;   only the interrupt my change this bit

; BIT_O_LOCK_SW is set if the interrupt is requested to switch the buffers
;   if this bit is set, both buffers are owned by the interrupt and writing is forbidden
;   this bit is set by non-interrupt sowftware to request a buffer switch
;   this bit is cleared by the interrupt after the buffer switch is done

; greyscale mapping for the greyscale value
; - contains values to compare swapped greyscale values to
;   - swapping makes it possible to compare without and 0x0F
;   - must contain values with low nibble 0x0 for this to work
; - contains 0x10, 0x20, ..., 0xF0 for linear greyscales
O_GREY_MAP	res	15

; timing information for the different splashes (i.e. the grayscale steps)
; - contains values for timer 2 period register for different splashes
; - timer 2 is configuration: prescaler 1:4, postscaler 1:1
; - valid values: 49, ..., 99 (i.e. timer 2 period 200, ..., 400 cycles)
O_TIMING	res	15



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; output initialization
;-------------------------------------------------------------------------------------------
O_INIT:
; initialize variables for LED output
	clrf	O_GREY			; first greyscale
	movff	O_GREY_MAP+0,O_GREY_VAL
	clrf	O_LINE			; next line is line 0
	clrf	O_LINE0			; first line bit active
	clrf	O_LINE1
	clrf	O_LINE2
	bsf	O_LINE0,0
	bcf	BIT_O_INT_BUF		; buffer 0 owned by interrupt
	bcf	BIT_O_BUF_SW		; no buffer switch requested
	lfsr	FSR2,O_FRAME0		; set pointer to line 0 of buffer 0
; initialize output style
	movlw	O_STYLE_LED		; use LED-optimized output style
	movwf	O_STYLE
; initialize timing information
	movlw	15
	movwf	A1_REG
	lfsr	FSR0,O_TIMING
O_INIT_TIMING:
	movlw	74			; (74+1)*4 = 300 cylces
	movwf	POSTINC0
	decfsz	A1_REG
	bra	O_INIT_TIMING
; done
	return

;-------------------------------------------------------------------------------------------
; timer 2 interrupt
; - called every 300 cycles
; - high priority
; - uses built in shadow registers for WREG, STATUS and BSR
; - uses FSR2, its value must not be changed anywhere else in the program
;-------------------------------------------------------------------------------------------
O_INT_TIMER2:
	bcf	PIR1,1			; clear interrupt flag of timer 2 interrupt

; --- int_grey_map --- start --- 3 or 18 cycles ---

; getting greyscale compare value is only needed in first line of splash
	movf	O_LINE,W
	bnz	O_INT_TIMER2_GREY_VAL_END
; get greyscale compare value for current greyscale value
	movff	FSR2L,O_FSR_BAK+0	; save FSR2
	movff	FSR2H,O_FSR_BAK+1
	lfsr	FSR2,O_GREY_MAP		; start address of greyscale mapping buffer
	movf	O_GREY,W		; add offset
	addwf	FSR2L
	movlw	0
	addwfc	FSR2H
	movff	INDF2,O_GREY_VAL	; save greyscale compare value
	movff	O_FSR_BAK+0,FSR2L	; restore FSR2
	movff	O_FSR_BAK+1,FSR2H
O_INT_TIMER2_GREY_VAL_END:

; --- int_grey_map --- end --- 3 or 18 cycles ---

; --- int_jump --- start --- 4, 7 or 8 cycles ---

; jump to selected output procedure
	movlw	O_STYLE_LATCH		; jump if latch-optimized output requested
	xorwf	O_STYLE,W
	bz	O_INT_TIMER2_LATCH
	movlw	O_STYLE_AMP		; jump if amplifier-optimized output requested
	xorwf	O_STYLE,W
	btfsc	STATUS,Z
	bra	O_INT_TIMER2_AMP

; --- int_jump --- end --- 4, 7 or 8 cycles ---

; output line in LED-optimized style

; --- line_out_leds --- start --- 87, 99 or 100 cycles ---

; all LEDs off by default
	clrf	O_LED0
	clrf	O_LED1
	clrf	O_LED2
	clrf	O_LED3

; LED 0
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,0		; LED 0 on
; LED 1
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,1		; LED 1 on
; LED 2
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,2		; LED 2 on
; LED 3
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,3		; LED 3 on
; LED 4
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,4		; LED 4 on
; LED 5
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,5		; LED 5 on
; LED 6
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,6		; LED 6 on
; LED 7
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,7		; LED 7 on

; LED 8
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,0		; LED 8 on
; LED 9
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,1		; LED 9 on
; LED 10
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,2		; LED 10 on
; LED 11
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,3		; LED 11 on
; LED 12
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,4		; LED 12 on

; LED 13
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,5		; LED 13 on
; LED 14
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,6		; LED 14 on
; LED 15
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,7		; LED 15 on
; LED 16
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,0		; LED 16 on
; LED 17
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,1		; LED 17 on
; LED 18
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,2		; LED 18 on
; LED 19
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,3		; LED 19 on
; LED 20
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,4		; LED 20 on
; LED 21
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,5		; LED 21 on
; LED 22
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,6		; LED 22 on
; LED 23
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,7		; LED 23 on
; LED 24
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED3,4		; LED 24 on
; LED 25
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED3,5		; LED 25 on

; clear all line and column outputs
	clrf	PORT_ROW0
	clrf	PORT_ROW1
	clrf	PORT_ROW2_COL3		; bits 0..3 rows, bits 5..6 columns, bits 6..7 not existent
	clrf	PORT_COL0
	clrf	PORT_COL1
	clrf	PORT_COL2

; check if power supply is detected
	btfsc	PIN_SH_A_PWR		; skip if no power supply is detected
	bra	O_INT_TIMER2_LEDS_ON	; jump to real output if power supply detected
	btfss	BIT_USB_PWR_OK		; skip if no usb power is allowed
	bra	O_INT_TIMER2_LEDS_OFF	; no power supply and no usb power allowed
O_INT_TIMER2_LEDS_ON:

; output column bits and O_LINE bits to LEDs
	movff	O_LED0,PORT_COL0
	movff	O_LED1,PORT_COL1
	movff	O_LED2,PORT_COL2
	movf	O_LED3,W		; bits 0..3,6..7 always 0
	iorwf	O_LINE2,W		; bits 4..7 always 0
	movwf	PORT_ROW2_COL3
	movff	O_LINE0,PORT_ROW0
	movff	O_LINE1,PORT_ROW1

O_INT_TIMER2_LEDS_OFF:

; --- line_out_leds --- end --- 87, 99 or 100 cycles ---

	bra	O_INT_TIMER2_TIMING

; output line in latch-optimized style
O_INT_TIMER2_LATCH:

; --- line_out_latch --- start --- 90, 103 or 104 cycles ---

; clear all line outputs
	clrf	PORT_ROW0
	clrf	PORT_ROW1
	movlw	0x30			; bits 0..3 rows, bits 6..7 not existent
	andwf	PORT_ROW2_COL3

; all lights off by default
	clrf	O_LED0
	clrf	O_LED1
	clrf	O_LED2
	clrf	O_LED3

; light 0
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,0		; light 0 on
; light 1
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,1		; light 1 on
; light 2
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,2		; light 2 on
; light 3
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,3		; light 3 on
; light 4
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,4		; light 4 on
; light 5
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,5		; light 5 on
; light 6
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,6		; light 6 on
; light 7
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED0,7		; light 7 on

; light 8
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,0		; light 8 on
; light 9
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,1		; light 9 on
; light 10
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,2		; light 10 on
; light 11
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,3		; light 11 on
; light 12
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,4		; light 12 on

; light 13
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,5		; light 13 on
; light 14
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,6		; light 14 on
; light 15
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED1,7		; light 15 on
; light 16
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,0		; light 16 on
; light 17
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,1		; light 17 on
; light 18
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,2		; light 18 on
; light 19
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,3		; light 19 on
; light 20
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,4		; light 20 on
; light 21
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,5		; light 21 on
; light 22
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,6		; light 22 on
; light 23
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED2,7		; light 23 on
; light 24
	movf	INDF2,W			; get greyscale value of first light in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED3,4		; light 24 on
; light 25
	swapf	POSTINC2,W		; get greyscale value of second light in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of light with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on light (skip next instruction)
	bsf	O_LED3,5		; light 25 on

; clear column outputs
	clrf	PORT_COL0
	clrf	PORT_COL1
	clrf	PORT_COL2
	movlw	0x0F			; bits 5..6 columns, bits 6..7 not existent
	andwf	PORT_ROW2_COL3

; check if power supply is detected
	btfsc	PIN_SH_A_PWR		; skip if no power supply is detected
	bra	O_INT_TIMER2_LATCH_ON	; jump to real output if power supply detected
	btfss	BIT_USB_PWR_OK		; skip if usb power is allowed
	bra	O_INT_TIMER2_LATCH_OFF	; no power supply and no usb power allowed
O_INT_TIMER2_LATCH_ON:

; output column bits to data pins
	movff	O_LED0,PORT_COL0
	movff	O_LED1,PORT_COL1
	movff	O_LED2,PORT_COL2
	movff	O_LED3,PORT_ROW2_COL3	; bits 0..3,6..7 always 0

; this separation should be enough for latch-circuits like ARCADEmaxi,
; because the ULN2803 in the row wires have got a much higher propagation
; delay than the 74HC541 in the column wires

; output O_LINE bits to latch enable pins
	movff	O_LINE0,PORT_ROW0
	movff	O_LINE1,PORT_ROW1
	movf	O_LINE2,W		; bits 4..7 always 0
	iorwf	PORT_ROW2_COL3

O_INT_TIMER2_LATCH_OFF:

; --- line_out_latch --- end --- 90, 103 or 104 cycles ---

	bra	O_INT_TIMER2_TIMING

; output line in amplifier-optimized style
O_INT_TIMER2_AMP:

; --- line_out_amp --- start --- 87, 99 or 100 cycles ---

; all LEDs off by default
	clrf	O_LED0
	clrf	O_LED1
	clrf	O_LED2
	clrf	O_LED3

; LED 0 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,0		; LED 0 on
; LED 1 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,1		; LED 1 on
; LED 2 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,2		; LED 2 on
; LED 3 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,3		; LED 3 on
; LED 4 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,4		; LED 4 on
; LED 5 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,5		; LED 5 on
; LED 6 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,6		; LED 6 on
; LED 7 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED0,7		; LED 7 on

; LED 8 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,0		; LED 8 on
; LED 9 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,1		; LED 9 on
; LED 10 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,2		; LED 10 on
; LED 11 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,3		; LED 11 on
; LED 12 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,4		; LED 12 on

; clear all line and column outputs
; (do this already here to give amplifier circuit time to switch off the current)
	clrf	PORT_ROW0
	clrf	PORT_ROW1
	clrf	PORT_ROW2_COL3		; bits 0..3 rows, bits 5..6 columns, bits 6..7 not existent
	clrf	PORT_COL0
	clrf	PORT_COL1
	clrf	PORT_COL2

; LED 13 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,5		; LED 13 on
; LED 14 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,6		; LED 14 on
; LED 15 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED1,7		; LED 15 on
; LED 16 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,0		; LED 16 on
; LED 17 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,1		; LED 17 on
; LED 18 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,2		; LED 18 on
; LED 19 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,3		; LED 19 on
; LED 20 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,4		; LED 20 on
; LED 21 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,5		; LED 21 on
; LED 22 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,6		; LED 22 on
; LED 23 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED2,7		; LED 23 on
; LED 24 (amplified)
	movf	INDF2,W			; get greyscale value of first LED in byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED3,4		; LED 24 on
; LED 25 (amplified)
	swapf	POSTINC2,W		; get greyscale value of second LED in byte
					; and set FSR0 to next byte
	cpfsgt	O_GREY_VAL		; compare greyscale value of LED with current counter
					; if counter (O_GREY_VAL) is greater than pixel (W)
					; then do _not_ turn on LED (skip next instruction)
	bsf	O_LED3,5		; LED 25 on

; check if power supply is detected
	btfsc	PIN_SH_A_PWR		; skip if no power supply is detected
	bra	O_INT_TIMER2_AMP_ON	; jump to real output if power supply detected
	btfss	BIT_USB_PWR_OK		; skip if no usb power is allowed
	bra	O_INT_TIMER2_AMP_OFF	; no power supply and no usb power allowed
O_INT_TIMER2_AMP_ON:

; output column bits and O_LINE bits to LEDs
	movff	O_LED0,PORT_COL0
	movff	O_LED1,PORT_COL1
	movff	O_LED2,PORT_COL2
	movf	O_LED3,W		; bits 0..3,6..7 always 0
	iorwf	O_LINE2,W		; bits 4..7 always 0
	movwf	PORT_ROW2_COL3
	movff	O_LINE0,PORT_ROW0
	movff	O_LINE1,PORT_ROW1

O_INT_TIMER2_AMP_OFF:

; --- line_out_amp --- end --- 87, 99 or 100 cycles ---

O_INT_TIMER2_TIMING:

; --- int_greyval_time --- start --- 3 or 18 cycles ---

; setting timing is only needed at first line of splash
	movf	O_LINE,W
	bnz	O_INT_TIMER2_TIMING_END
; get timing information for current greyscale value
	movff	FSR2L,O_FSR_BAK+0	; save FSR2
	movff	FSR2H,O_FSR_BAK+1
	lfsr	FSR2,O_TIMING		; start address of timing information buffer
	movf	O_GREY,W		; add offset
	addwf	FSR2L
	movlw	0
	addwfc	FSR2H
	movff	INDF2,PR2		; put timing value into timer 2 period register
	movff	O_FSR_BAK+0,FSR2L	; restore FSR2
	movff	O_FSR_BAK+1,FSR2H
O_INT_TIMER2_TIMING_END:

; --- int_greyval_time --- end --- 3 or 18 cycles ---

; --- next_line --- start --- 9, 21, 24 or 26 cycles ---

; next line
	incf	O_LINE			; next line
	bcf	STATUS,C		; calculate new line bits
	rlcf	O_LINE0
	rlcf	O_LINE1
	rlcf	O_LINE2
; last line?
	movlw	19
	cpfsgt	O_LINE			; skip if O_LINE > 19
	bra	O_INT_TIMER2_END
; first line
	clrf	O_LINE			; next line is line 0
	clrf	O_LINE0			; first line bit active
	clrf	O_LINE1
	clrf	O_LINE2
	bsf	O_LINE0,0
; set pointer to line 0 of buffer owned by interrupt
	lfsr	FSR2,O_FRAME0
	btfsc	BIT_O_INT_BUF		; skip if buffer 0 is owned by interrupt
	lfsr	FSR2,O_FRAME1
; use next greyscale value for next splash
	incf	O_GREY
; last splash (i.e. last greyscale)?
	movlw	14
	cpfsgt	O_GREY			; skip if O_GREY > 14
	bra	O_INT_TIMER2_END
; first splash
	clrf	O_GREY			; first greyscale
; check if buffer switch requested
	btfss	BIT_O_BUF_SW		; skip if buffer switch requested
	bra	O_INT_TIMER2_END
; switch buffers
	movlw	BIT_O_INT_BUF_MASK	; complement BIT_O_INT_BUF
	xorwf	BIT_O_INT_BUF_BYTE
	bcf	BIT_O_BUF_SW		; buffer switch is done	

O_INT_TIMER2_END:

; --- next_line --- end --- 9, 21, 24 or 26 cycles ---

	retfie	FAST



;===========================================================================================
; global variables
;===========================================================================================
	global		O_FRAME0, O_FRAME1, O_STYLE, O_GREY_MAP, O_TIMING



;===========================================================================================
; global procedures
;===========================================================================================
	global		O_INIT, O_INT_TIMER2



	end
